# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: api.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


import diagonal_b6.geometry_pb2 as geometry__pb2
import diagonal_b6.features_pb2 as features__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\tapi.proto\x12\x03\x61pi\x1a\x0egeometry.proto\x1a\x0e\x66\x65\x61tures.proto\"]\n\x0f\x43ollectionProto\x12#\n\x04keys\x18\x02 \x03(\x0b\x32\x15.api.LiteralNodeProto\x12%\n\x06values\x18\x04 \x03(\x0b\x32\x15.api.LiteralNodeProto\"X\n\tPairProto\x12$\n\x05\x66irst\x18\x01 \x01(\x0b\x32\x15.api.LiteralNodeProto\x12%\n\x06second\x18\x02 \x01(\x0b\x32\x15.api.LiteralNodeProto\">\n\x15ModifiedFeaturesProto\x12%\n\x03ids\x18\x01 \x03(\x0b\x32\x18.features.FeatureIDProto\"l\n\x12\x41ppliedChangeProto\x12*\n\x08original\x18\x01 \x03(\x0b\x32\x18.features.FeatureIDProto\x12*\n\x08modified\x18\x02 \x03(\x0b\x32\x18.features.FeatureIDProto\"\xb8\x01\n\tNodeProto\x12\x10\n\x06symbol\x18\x01 \x01(\tH\x00\x12(\n\x07literal\x18\x02 \x01(\x0b\x32\x15.api.LiteralNodeProtoH\x00\x12\"\n\x04\x63\x61ll\x18\x03 \x01(\x0b\x32\x12.api.CallNodeProtoH\x00\x12\'\n\x07lambda_\x18\x04 \x01(\x0b\x32\x14.api.LambdaNodeProtoH\x00\x12\r\n\x05\x42\x65gin\x18\x06 \x01(\x05\x12\x0b\n\x03\x45nd\x18\x07 \x01(\x05\x42\x06\n\x04node\"\xe9\x04\n\x10LiteralNodeProto\x12\x12\n\x08nilValue\x18\x01 \x01(\x08H\x00\x12\x13\n\tboolValue\x18\x02 \x01(\x08H\x00\x12\x15\n\x0bstringValue\x18\x03 \x01(\tH\x00\x12\x12\n\x08intValue\x18\x04 \x01(\x03H\x00\x12\x14\n\nfloatValue\x18\x05 \x01(\x01H\x00\x12/\n\x0f\x63ollectionValue\x18\x06 \x01(\x0b\x32\x14.api.CollectionProtoH\x00\x12#\n\tpairValue\x18\x07 \x01(\x0b\x32\x0e.api.PairProtoH\x00\x12.\n\x0c\x66\x65\x61tureValue\x18\x08 \x01(\x0b\x32\x16.features.FeatureProtoH\x00\x12%\n\nqueryValue\x18\t \x01(\x0b\x32\x0f.api.QueryProtoH\x00\x12\x32\n\x0e\x66\x65\x61tureIDValue\x18\n \x01(\x0b\x32\x18.features.FeatureIDProtoH\x00\x12*\n\npointValue\x18\x0b \x01(\x0b\x32\x14.geometry.PointProtoH\x00\x12,\n\tpathValue\x18\x0c \x01(\x0b\x32\x17.geometry.PolylineProtoH\x00\x12\x30\n\tareaValue\x18\r \x01(\x0b\x32\x1b.geometry.MultiPolygonProtoH\x00\x12\x35\n\x12\x61ppliedChangeValue\x18\x0e \x01(\x0b\x32\x17.api.AppliedChangeProtoH\x00\x12\x16\n\x0cgeoJSONValue\x18\x0f \x01(\x0cH\x00\x12&\n\x08tagValue\x18\x10 \x01(\x0b\x32\x12.features.TagProtoH\x00\x42\x07\n\x05value\"O\n\rCallNodeProto\x12 \n\x08\x66unction\x18\x01 \x01(\x0b\x32\x0e.api.NodeProto\x12\x1c\n\x04\x61rgs\x18\x02 \x03(\x0b\x32\x0e.api.NodeProto\"=\n\x0fLambdaNodeProto\x12\x0c\n\x04\x61rgs\x18\x01 \x03(\t\x12\x1c\n\x04node\x18\x02 \x01(\x0b\x32\x0e.api.NodeProto\"\x1c\n\rKeyQueryProto\x12\x0b\n\x03key\x18\x01 \x01(\t\"0\n\x12KeyValueQueryProto\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t\"V\n\x0fTypedQueryProto\x12#\n\x04type\x18\x01 \x01(\x0e\x32\x15.features.FeatureType\x12\x1e\n\x05query\x18\x02 \x01(\x0b\x32\x0f.api.QueryProto\"0\n\x0cQueriesProto\x12 \n\x07queries\x18\x01 \x03(\x0b\x32\x0f.api.QueryProto\"\x0f\n\rAllQueryProto\"\x11\n\x0f\x45mptyQueryProto\"F\n\x08\x43\x61pProto\x12$\n\x06\x63\x65nter\x18\x01 \x01(\x0b\x32\x14.geometry.PointProto\x12\x14\n\x0cradiusMeters\x18\x02 \x01(\x01\"#\n\x0eS2CellIDsProto\x12\x11\n\ts2CellIDs\x18\x01 \x03(\x04\"\xf1\x04\n\nQueryProto\x12!\n\x03\x61ll\x18\x01 \x01(\x0b\x32\x12.api.AllQueryProtoH\x00\x12%\n\x05\x65mpty\x18\x02 \x01(\x0b\x32\x14.api.EmptyQueryProtoH\x00\x12\x0f\n\x05keyed\x18\x03 \x01(\tH\x00\x12$\n\x06tagged\x18\x04 \x01(\x0b\x32\x12.features.TagProtoH\x00\x12%\n\x05typed\x18\x05 \x01(\x0b\x32\x14.api.TypedQueryProtoH\x00\x12)\n\x0cintersection\x18\x06 \x01(\x0b\x32\x11.api.QueriesProtoH\x00\x12\"\n\x05union\x18\x07 \x01(\x0b\x32\x11.api.QueriesProtoH\x00\x12&\n\rintersectsCap\x18\x08 \x01(\x0b\x32\r.api.CapProtoH\x00\x12\x35\n\x11intersectsFeature\x18\t \x01(\x0b\x32\x18.features.FeatureIDProtoH\x00\x12/\n\x0fintersectsPoint\x18\n \x01(\x0b\x32\x14.geometry.PointProtoH\x00\x12\x35\n\x12intersectsPolyline\x18\x0b \x01(\x0b\x32\x17.geometry.PolylineProtoH\x00\x12=\n\x16intersectsMultiPolygon\x18\x0c \x01(\x0b\x32\x1b.geometry.MultiPolygonProtoH\x00\x12.\n\x0fintersectsCells\x18\r \x01(\x0b\x32\x13.api.S2CellIDsProtoH\x00\x12-\n\x0emightIntersect\x18\x0e \x01(\x0b\x32\x13.api.S2CellIDsProtoH\x00\x42\x07\n\x05query\"C\n\x1b\x46indFeatureByIDRequestProto\x12$\n\x02id\x18\x01 \x01(\x0b\x32\x18.features.FeatureIDProto\"G\n\x1c\x46indFeatureByIDResponseProto\x12\'\n\x07\x66\x65\x61ture\x18\x01 \x01(\x0b\x32\x16.features.FeatureProto\":\n\x18\x46indFeaturesRequestProto\x12\x1e\n\x05query\x18\x01 \x01(\x0b\x32\x0f.api.QueryProto\"E\n\x19\x46indFeaturesResponseProto\x12(\n\x08\x66\x65\x61tures\x18\x01 \x03(\x0b\x32\x16.features.FeatureProto\"`\n\x16ModifyTagsRequestProto\x12$\n\x02id\x18\x01 \x01(\x0b\x32\x18.features.FeatureIDProto\x12 \n\x04tags\x18\x02 \x03(\x0b\x32\x12.features.TagProto\"L\n\x1bModifyTagsBatchRequestProto\x12-\n\x08requests\x18\x01 \x03(\x0b\x32\x1b.api.ModifyTagsRequestProto\"\x1e\n\x1cModifyTagsBatchResponseProto\"H\n\x14\x45valuateRequestProto\x12\x1f\n\x07request\x18\x01 \x01(\x0b\x32\x0e.api.NodeProto\x12\x0f\n\x07version\x18\x02 \x01(\t\"7\n\x15\x45valuateResponseProto\x12\x1e\n\x06result\x18\x01 \x01(\x0b\x32\x0e.api.NodeProto2G\n\x02\x42\x36\x12\x41\n\x08\x45valuate\x12\x19.api.EvaluateRequestProto\x1a\x1a.api.EvaluateResponseProtoB\x19Z\x17\x64iagonal.works/b6/protob\x06proto3')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'api_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'Z\027diagonal.works/b6/proto'
  _COLLECTIONPROTO._serialized_start=50
  _COLLECTIONPROTO._serialized_end=143
  _PAIRPROTO._serialized_start=145
  _PAIRPROTO._serialized_end=233
  _MODIFIEDFEATURESPROTO._serialized_start=235
  _MODIFIEDFEATURESPROTO._serialized_end=297
  _APPLIEDCHANGEPROTO._serialized_start=299
  _APPLIEDCHANGEPROTO._serialized_end=407
  _NODEPROTO._serialized_start=410
  _NODEPROTO._serialized_end=594
  _LITERALNODEPROTO._serialized_start=597
  _LITERALNODEPROTO._serialized_end=1214
  _CALLNODEPROTO._serialized_start=1216
  _CALLNODEPROTO._serialized_end=1295
  _LAMBDANODEPROTO._serialized_start=1297
  _LAMBDANODEPROTO._serialized_end=1358
  _KEYQUERYPROTO._serialized_start=1360
  _KEYQUERYPROTO._serialized_end=1388
  _KEYVALUEQUERYPROTO._serialized_start=1390
  _KEYVALUEQUERYPROTO._serialized_end=1438
  _TYPEDQUERYPROTO._serialized_start=1440
  _TYPEDQUERYPROTO._serialized_end=1526
  _QUERIESPROTO._serialized_start=1528
  _QUERIESPROTO._serialized_end=1576
  _ALLQUERYPROTO._serialized_start=1578
  _ALLQUERYPROTO._serialized_end=1593
  _EMPTYQUERYPROTO._serialized_start=1595
  _EMPTYQUERYPROTO._serialized_end=1612
  _CAPPROTO._serialized_start=1614
  _CAPPROTO._serialized_end=1684
  _S2CELLIDSPROTO._serialized_start=1686
  _S2CELLIDSPROTO._serialized_end=1721
  _QUERYPROTO._serialized_start=1724
  _QUERYPROTO._serialized_end=2349
  _FINDFEATUREBYIDREQUESTPROTO._serialized_start=2351
  _FINDFEATUREBYIDREQUESTPROTO._serialized_end=2418
  _FINDFEATUREBYIDRESPONSEPROTO._serialized_start=2420
  _FINDFEATUREBYIDRESPONSEPROTO._serialized_end=2491
  _FINDFEATURESREQUESTPROTO._serialized_start=2493
  _FINDFEATURESREQUESTPROTO._serialized_end=2551
  _FINDFEATURESRESPONSEPROTO._serialized_start=2553
  _FINDFEATURESRESPONSEPROTO._serialized_end=2622
  _MODIFYTAGSREQUESTPROTO._serialized_start=2624
  _MODIFYTAGSREQUESTPROTO._serialized_end=2720
  _MODIFYTAGSBATCHREQUESTPROTO._serialized_start=2722
  _MODIFYTAGSBATCHREQUESTPROTO._serialized_end=2798
  _MODIFYTAGSBATCHRESPONSEPROTO._serialized_start=2800
  _MODIFYTAGSBATCHRESPONSEPROTO._serialized_end=2830
  _EVALUATEREQUESTPROTO._serialized_start=2832
  _EVALUATEREQUESTPROTO._serialized_end=2904
  _EVALUATERESPONSEPROTO._serialized_start=2906
  _EVALUATERESPONSEPROTO._serialized_end=2961
  _B6._serialized_start=2963
  _B6._serialized_end=3034
# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: geometry.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x0egeometry.proto\x12\x08geometry\"L\n\rPolylineProto\x12$\n\x06points\x18\x01 \x03(\x0b\x32\x14.geometry.PointProto\x12\x15\n\rlength_meters\x18\x02 \x01(\x01\"=\n\x11MultiPolygonProto\x12(\n\x08polygons\x18\x01 \x03(\x0b\x32\x16.geometry.PolygonProto\"2\n\x0cPolygonProto\x12\"\n\x05loops\x18\x01 \x03(\x0b\x32\x13.geometry.LoopProto\"1\n\tLoopProto\x12$\n\x06points\x18\x01 \x03(\x0b\x32\x14.geometry.PointProto\",\n\nPointProto\x12\x0e\n\x06lat_e7\x18\x01 \x01(\x05\x12\x0e\n\x06lng_e7\x18\x02 \x01(\x05\x42\x19Z\x17\x64iagonal.works/b6/protob\x06proto3')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'geometry_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'Z\027diagonal.works/b6/proto'
  _POLYLINEPROTO._serialized_start=28
  _POLYLINEPROTO._serialized_end=104
  _MULTIPOLYGONPROTO._serialized_start=106
  _MULTIPOLYGONPROTO._serialized_end=167
  _POLYGONPROTO._serialized_start=169
  _POLYGONPROTO._serialized_end=219
  _LOOPPROTO._serialized_start=221
  _LOOPPROTO._serialized_end=270
  _POINTPROTO._serialized_start=272
  _POINTPROTO._serialized_end=316
# @@protoc_insertion_point(module_scope)



import grpc

from diagonal_b6 import expression
from diagonal_b6 import VERSION

from diagonal_b6 import api_pb2
from diagonal_b6 import api_pb2_grpc

class Connection:

    def __init__(self, stub):
        self.stub = stub

    def __call__(self, e):
        request = api_pb2.EvaluateRequestProto()
        request.version = VERSION
        request.request.CopyFrom(expression.to_node_proto(e))
        return expression.from_node_proto(self.stub.Evaluate(request).result)

def connect_insecure(address):
    channel = grpc.insecure_channel(address)
    return Connection(api_pb2_grpc.B6Stub(channel))

def connect(address, token, root_certificates=None):
    channel_credentials = grpc.ssl_channel_credentials(root_certificates=root_certificates)
    credentials = grpc.composite_channel_credentials(channel_credentials, grpc.access_token_call_credentials(token))
    channel = grpc.secure_channel(address, credentials)
    return Connection(api_pb2_grpc.B6Stub(channel))



#!/usr/bin/env python3

import sys
import json

from copy import copy
from datetime import datetime

def name_for_function(name):
    if name in ["or", "and"]:
        return name + "_"
    return name.replace("-", "_")

def name_for_type(t):
    if t == "string":
        return "String"
    elif t == "int":
        return "Int"
    elif t == "float64":
        return "Float"
    elif t == "any":
        return "Any"
    elif t == "bool":
        return "Bool"
    return t

def name_for_traits(t):
    return "%sTraits" % name_for_type(t)

def name_for_collection_values(t):
    return "%sValues" % name_for_type(t)

def name_for_result(t):
    return "%sResult" % name_for_type(t)

def name_for_any_key_result(t):
    return "Any%sResult" % name_for_type(t)

def name_for_collection_of_traits(t, collections):
    for (name, (key, value)) in collections.items():
        if key == "any" and value == t:
            return name_for_traits(name)
    return "CollectionTraits"

def name_for_collection_of_result(t, collections):
    for (name, (key, value)) in collections.items():
        if key == "any" and value == t:
            return name_for_result(name)
    return "CollectionResult"

def output_traits(t, functions, collections, hints, parents):
    if len(parents.get(t, [])) > 0:
        print("class %s(%s):" % (name_for_traits(t), ", ".join([name_for_traits(p) for p in parents[t]])))
    else:
        print("class %s:" % name_for_traits(t))
    methods = 0
    for f in functions:            
        if len(f["Args"]) > 0 and f["Args"][0] == t:
            if methods == 0:
                print("")
            signature = ", ".join(["self"] + ["a%d: %s" % (i, hints[a]) for (i, a) in enumerate(f["Args"][1:])])
            print("    def %s(%s) -> %s:" % (name_for_function(f["Name"]), signature, hints[f["Result"]]))
            args = ", ".join(["self"] + ["a%d" % i for i in range(0, len(f["Args"][1:]))])
            print("        return %s(%s)" % (name_for_function(f["Name"]), args))
            print("")
            methods += 1
    print("    @classmethod")
    print("    def _collection(cls):")
    print("        return %s" % (name_for_collection_of_result(t, collections),))
    print("")

def output_collection_values_traits(t, functions, collections, hints, parents):
    n = name_for_collection_values(t)
    if len(parents.get(t, [])) > 0:
        print("class %s(%s):" % (name_for_traits(n), ", ".join([name_for_traits(name_for_collection_values(p)) for p in parents[t]])))
    else:
        print("class %s:" % name_for_traits(n))
    methods = 0
    for f in functions:            
        if len(f["Args"]) > 0 and f["Args"][0] == t:
            if methods == 0:
                print("")
            signature = ", ".join(["self"] + ["a%d: %s" % (i, hints[a]) for (i, a) in enumerate(f["Args"][1:])])
            print("    def %s(%s) -> %s:" % (name_for_function(f["Name"]), signature, name_for_collection_of_traits(f["Result"], collections)))
            args = ", ".join(["a%d" % i for i in range(0, len(f["Args"][1:]))])
            if len(args) > 0:
                print("        return self.map(Lambda(lambda x: %s(x, %s), [self._values()]))" % (name_for_function(f["Name"]), args))
            else:
                print("        return self.map(Lambda(%s, [self._values()]))" % name_for_function(f["Name"]))
            print("")
            methods += 1
    if methods == 0:
        print("    pass")
        print("")
    
    print("class %s(Result, %s, CollectionTraits):" % (name_for_any_key_result(n), name_for_traits(n)))
    print("")
    print("    def __init__(self, node):")
    print("        Result.__init__(self, node)")
    print("")
    print("    @classmethod")
    print("    def _values(cls):")
    print("        return %s" % (name_for_result(t)))
    print("")

def output_function_arg_result(t, hints):
    print("class %s(Result, %s):" % (name_for_result(t["Name"]), hints[t["Name"]],))
    print("")
    print("    def __init__(self, node):")
    print("        Result.__init__(self, node)")
    print("")
    args = ", ".join(["a%d : %s" % (i, hints[at]) for (i, at) in enumerate(t["Args"])])
    print("    def __call__(self, %s) -> %s:" % (args, hints[t["Result"]]))
    print("        raise NotImplementedError()")
    print("")

SPECIAL_FUNCTIONS = ("map", "filter")

COLLECTION_PARENTS = {
    "PointFeatureCollection": "PointCollection",
    "PathFeatureCollection": "PathCollection",
    "AreaFeatureCollection": "AreaCollection",
}

def ancestors(t, parents):
    queue = copy(parents.get(t, []))
    ancestors = []
    while len(queue) > 0:
        ancestors.append(queue.pop())
        queue.extend(parents.get(ancestors[-1], []))
    return ancestors

def main():
    api = json.load(sys.stdin)
    print("# Code generated by generate_api.py. DO NOT EDIT.")
    print("# Client library for Diagonal's geospatial analysis engine, b6.")
    print("")
    print("from __future__ import annotations")
    print("")
    print("from typing import Callable")
    print("")
    print("import diagonal_b6.expression")
    print("from diagonal_b6.expression import Call, Symbol, Lambda, Result")
    print("")
    print("VERSION = %s" % repr(api["Version"]))
    print("")

    traits = set()
    parents = {}
    hints = {}
    for t in ("any", "int", "float64", "bool", "string"):
        traits.add(t)
        parents[t] = []
        hints[t] = name_for_traits(t)
    for t in ("int", "float64"):
        parents[t].append("Number")
    for t in api["Interfaces"]:
        traits.add(t["Name"])
        hints[t["Name"]] = name_for_traits(t["Name"])
        parents[t["Name"]] = t["Implements"]
        for tt in t["Implements"]:
            traits.add(tt)
            hints[tt] = name_for_traits(tt)

    reference_counts = {}
    collections = {}
    collection_values = set()
    for t in api["Collections"]:
        traits.add(t["Name"])
        traits.add(t["Value"])
        collections[t["Name"]] = (t["Key"], t["Value"])
        collection_values.add(t["Value"])
        for n in (t["Value"], "Collection"):
            reference_counts[n] = reference_counts.get(n, 0) + 1
        for a in ancestors(t["Value"], parents):
            collection_values.add(a)
            reference_counts[a] = reference_counts.get(a, 0) + 1
        hints[t["Name"]] = name_for_result(t["Name"])
        if t["Name"] != "Collection":
            if t["Name"] in COLLECTION_PARENTS:
                parents[t["Name"]] = [COLLECTION_PARENTS[t["Name"]], name_for_collection_values(t["Value"])]
            else:
                parents[t["Name"]] = ["Collection", name_for_collection_values(t["Value"])]
    for t in api["FunctionArgs"]:
        hints[t["Name"]] = "Callable[[%s],%s]" % (",".join([hints[a] for a in t["Args"]]), hints[t["Result"]])

    for t in traits:
        for a in ancestors(t, parents):
            reference_counts[a] = reference_counts.get(a, 0) + 1
    traits = list(traits)
    traits.sort(key=lambda t: -reference_counts.get(t, 0))

    for t in traits:
        output_traits(t, api["Functions"], collections, hints, parents)
        if t in collection_values:
            output_collection_values_traits(t, api["Functions"], collections, hints, parents)

    for t in traits:
        print("class %s(Result, %s):" % (name_for_result(t), name_for_traits(t)))
        print("    def __init__(self, node):")
        print("        Result.__init__(self, node)")        
        print("")
        if t in collections:
            _, values = collections[t]
            print("    @classmethod")
            print("    def _values(cls):")
            print("        return %s" % name_for_result(values))
            print("")

    for t in api["FunctionArgs"]:
        output_function_arg_result(t, hints)

    for f in api["Functions"]:
        if f["Name"] in SPECIAL_FUNCTIONS:
            print("%s = diagonal_b6.expression._%s" % (f["Name"], f["Name"]))
        else:
            signature = ", ".join(["a%d: %s" % (i, hints[a]) for (i, a) in enumerate(f["Args"])])
            print("def %s(%s) -> %s:" % (name_for_function(f["Name"]), signature, hints[f["Result"]]))
            args = "[" + ", ".join(["a%d" % i for i in range(0, len(f["Args"]))])+ "]"
            print("    return %s(Call(Symbol(%s), %s))" % (name_for_result(f["Result"]), repr(f["Name"]), args))
        print("")

if __name__ == "__main__":
    main()


from diagonal_b6 import expression

from diagonal_b6 import api_pb2

class Query(expression.Literal):

    def __init__(self, query):
        self.query = query

    def tagged(self, key, value=None):        
        return self._and(tagged(key, value))

    def type(self, t):        
        return type(t, self)

    def within(self, area):        
        return self._and(within(area))

    def union(self, query):
        return union(self, query)

    def _and(self, other):
        if self.query.WhichOneof("query") == "all":
            return Query(other)
        elif self.query.WhichOneof("query") == "intersection":
            query = api_pb2.QueryProto()
            query.CopyFrom(self.query)
            query.intersection.queries.add().CopyFrom(other)
            return Query(query)
        else:
            query = api_pb2.QueryProto()
            query.intersection.queries.add().CopyFrom(self.query)
            query.intersection.queries.add().CopyFrom(other.query)
            return Query(query)

    def to_proto(self):
        return self.query

    def to_literal_proto(self):
        l = api_pb2.LiteralNodeProto()
        l.queryValue.CopyFrom(self.query)
        return l

def from_proto(pb):
    return Query(pb)

expression.register_literal("queryValue", from_proto)

def union(*queries):
    q = api_pb2.QueryProto()
    for query in queries:
        _flattern_union(q, query.query)
    return Query(q)

def _flattern_union(dest, source):
    if source.WhichOneof("query") == "union":
        for q in source.union.queries:
            _flattern_union(dest, q)
    else:
        dest.union.queries.add().CopyFrom(source)

def tagged(key, value=None):
    q = api_pb2.QueryProto()
    if value is None:
        q.key.key = key
    else:
        q.keyValue.key = key
        q.keyValue.value = value
    return Query(q)

def within(area):
    q = api_pb2.QueryProto()
    area._fill_query(q)
    return Query(q)

def type(t, query):
    q = api_pb2.QueryProto()
    q.type.type = t
    q.query.CopyFrom(query)
    return Query(q)

def all():
    q = api_pb2.QueryProto()
    q.all.CopyFrom(api_pb2.AllQueryProto())
    return Query(q)



import gzip
import inspect
import json

from diagonal_b6 import api_pb2

class Node:

    def to_node_proto(self):
        raise NotImplementedError()

class Literal(Node):

    def to_node_proto(self):
        node = api_pb2.NodeProto()
        node.literal.CopyFrom(self.to_literal_proto())
        return node

    def to_literal_proto(self):
        raise NotImplementedError()

class Symbol(Node):

    def __init__(self, name):
        self.name = name

    def to_node_proto(self):
        node = api_pb2.NodeProto()
        node.symbol = self.name
        return node

class Call(Node):

    def __init__(self, function, args):
        self.function = function
        self.args = args

    def to_node_proto(self):
        node = api_pb2.NodeProto()
        node.call.function.CopyFrom(self.function.to_node_proto())
        for arg in self.args:
            node.call.args.add().CopyFrom(to_node_proto(arg))
        return node

class Lambda(Node):

    def __init__(self, function, arg_types):
        self.function = function
        self.arg_types = arg_types

    def with_arg_types(self, arg_types):
        return Lambda(self.function, arg_types)

    def to_node_proto(self):
        n = api_pb2.NodeProto()
        args = ["_%s_%d" % (name, id(n)) for name in inspect.signature(self.function).parameters]
        for arg in args:
            n.lambda_.args.append(arg)
        n.lambda_.node.CopyFrom(to_node_proto(self.function(*[type(Symbol(name)) for (name, type) in zip(args, self.arg_types)])))
        return n

    def __call__(self, *args):
        return self.function(*args)

def to_lambda(f):
    if isinstance(f, Lambda):
        return f
    arg_types = []
    for p in inspect.signature(f).parameters.values():
        if p.annotation != inspect.Signature.empty:
            arg_types.append(p.annotation)
        else:
            arg_types.append(Result)
    return Lambda(f, arg_types)

class Result(Node):

    def __init__(self, node):
        self.node = node

    def to_node_proto(self):
        return self.node.to_node_proto()

class Placeholder(Node):

    def to_node_proto(self):
        raise Exception("Placeholder can't be converted to a proto")

def to_node_proto(v):
    if isinstance(v, Node):
        return v.to_node_proto()
    elif isinstance(v, int):
       n = api_pb2.NodeProto()
       n.literal.intValue = v
       return n
    elif isinstance(v, float):
        n = api_pb2.NodeProto()
        n.literal.floatValue = v
        return n
    elif isinstance(v, str):
        n = api_pb2.NodeProto()
        n.literal.stringValue = v
        return n
    elif callable(v):
        return to_lambda(v).to_node_proto()
    else:
        raise ValueError("Can't convert %s to proto" % (v,))

def from_node_proto(n):
    if n.WhichOneof("node") != "literal":
        raise Exception("Can't convert node type %s to a value" % (n.WhichOneof("node"),))
    return from_literal_node_proto(n.literal)

_literals = {}

def register_literal(oneof, from_proto):
    _literals[oneof] = from_proto

register_literal("nilValue", lambda v: None)
register_literal("intValue", lambda v: v)
register_literal("floatValue", lambda v: v)
register_literal("stringValue", lambda v: v)
register_literal("geoJSONValue", lambda v: json.loads(gzip.decompress(v)))

def from_literal_node_proto(literal):
    oneof = literal.WhichOneof("value")
    from_proto = _literals.get(oneof, None)
    if from_proto is not None:
        return from_proto(getattr(literal, oneof))
    else:
        raise Exception("can't convert %s to value" % (oneof,))

def from_collection_proto(collection):
    return [(from_literal_node_proto(collection.keys[i]), from_literal_node_proto(collection.values[i]))
            for i in range(0, len(collection.keys))]

register_literal("collectionValue", from_collection_proto)

def _map(collection, f):
    result = f(collection._values()(Placeholder()))
    return result._collection()(Call(Symbol("map"), [collection, to_lambda(f).with_arg_types((collection._values(),))]))

def _filter(collection, f):
    return type(collection)(Call(Symbol("filter"), [collection, to_lambda(f).with_arg_types((collection._values(),))]))





# Code generated by generate_api.py. DO NOT EDIT.
# Client library for Diagonal's geospatial analysis engine, b6.

from __future__ import annotations

from typing import Callable

import diagonal_b6.expression
from diagonal_b6.expression import Call, Symbol, Lambda, Result

VERSION = '0.0.2+modified'

class CollectionTraits:

    def take(self, a0: IntTraits) -> CollectionResult:
        return take(self, a0)

    def map_items(self, a0: Callable[[PairTraits],AnyTraits]) -> CollectionResult:
        return map_items(self, a0)

    def count_values(self) -> CollectionResult:
        return count_values(self)

    def map_parallel(self, a0: Callable[[AnyTraits],AnyTraits]) -> CollectionResult:
        return map_parallel(self, a0)

    def sum_by_key(self) -> CollectionResult:
        return sum_by_key(self)

    def count(self) -> IntTraits:
        return count(self)

    def filter(self, a0: Callable[[AnyTraits],BoolTraits]) -> CollectionResult:
        return filter(self, a0)

    def flattern(self) -> CollectionResult:
        return flattern(self)

    def map(self, a0: Callable[[AnyTraits],AnyTraits]) -> CollectionResult:
        return map(self, a0)

    @classmethod
    def _collection(cls):
        return CollectionResult

class RenderableTraits:

    def to_geojson(self) -> GeoJSONTraits:
        return to_geojson(self)

    @classmethod
    def _collection(cls):
        return AnyRenderableCollectionResult

class RenderableValuesTraits:

    def to_geojson(self) -> CollectionTraits:
        return self.map(Lambda(to_geojson, [self._values()]))

class AnyRenderableValuesResult(Result, RenderableValuesTraits, CollectionTraits):

    def __init__(self, node):
        Result.__init__(self, node)

    @classmethod
    def _values(cls):
        return RenderableResult

class GeometryTraits(RenderableTraits):

    def centroid(self) -> PointTraits:
        return centroid(self)

    def points(self) -> PointCollectionResult:
        return points(self)

    def intersecting(self) -> QueryTraits:
        return intersecting(self)

    def tile_paths(self, a0: IntTraits) -> IntStringCollectionResult:
        return tile_paths(self, a0)

    @classmethod
    def _collection(cls):
        return CollectionResult

class GeometryValuesTraits(RenderableValuesTraits):

    def centroid(self) -> PointCollectionTraits:
        return self.map(Lambda(centroid, [self._values()]))

    def points(self) -> CollectionTraits:
        return self.map(Lambda(points, [self._values()]))

    def intersecting(self) -> CollectionTraits:
        return self.map(Lambda(intersecting, [self._values()]))

    def tile_paths(self, a0: IntTraits) -> CollectionTraits:
        return self.map(Lambda(lambda x: tile_paths(x, a0), [self._values()]))

class AnyGeometryValuesResult(Result, GeometryValuesTraits, CollectionTraits):

    def __init__(self, node):
        Result.__init__(self, node)

    @classmethod
    def _values(cls):
        return GeometryResult

class IdentifiableTraits:

    def get_string(self, a0: StringTraits) -> StringTraits:
        return get_string(self, a0)

    def add_tag(self, a0: TagTraits) -> ChangeTraits:
        return add_tag(self, a0)

    def get(self, a0: StringTraits) -> TagTraits:
        return get(self, a0)

    def all_tags(self) -> IntTagCollectionResult:
        return all_tags(self)

    def debug_tokens(self) -> IntStringCollectionResult:
        return debug_tokens(self)

    def get_int(self, a0: StringTraits) -> IntTraits:
        return get_int(self, a0)

    def count_tag_value(self, a0: StringTraits) -> CollectionResult:
        return count_tag_value(self, a0)

    def get_float(self, a0: StringTraits) -> FloatTraits:
        return get_float(self, a0)

    def remove_tag(self, a0: StringTraits) -> ChangeTraits:
        return remove_tag(self, a0)

    @classmethod
    def _collection(cls):
        return CollectionResult

class IdentifiableValuesTraits:

    def get_string(self, a0: StringTraits) -> CollectionTraits:
        return self.map(Lambda(lambda x: get_string(x, a0), [self._values()]))

    def add_tag(self, a0: TagTraits) -> CollectionTraits:
        return self.map(Lambda(lambda x: add_tag(x, a0), [self._values()]))

    def get(self, a0: StringTraits) -> CollectionTraits:
        return self.map(Lambda(lambda x: get(x, a0), [self._values()]))

    def all_tags(self) -> CollectionTraits:
        return self.map(Lambda(all_tags, [self._values()]))

    def debug_tokens(self) -> CollectionTraits:
        return self.map(Lambda(debug_tokens, [self._values()]))

    def get_int(self, a0: StringTraits) -> CollectionTraits:
        return self.map(Lambda(lambda x: get_int(x, a0), [self._values()]))

    def count_tag_value(self, a0: StringTraits) -> CollectionTraits:
        return self.map(Lambda(lambda x: count_tag_value(x, a0), [self._values()]))

    def get_float(self, a0: StringTraits) -> AnyFloatCollectionTraits:
        return self.map(Lambda(lambda x: get_float(x, a0), [self._values()]))

    def remove_tag(self, a0: StringTraits) -> CollectionTraits:
        return self.map(Lambda(lambda x: remove_tag(x, a0), [self._values()]))

class AnyIdentifiableValuesResult(Result, IdentifiableValuesTraits, CollectionTraits):

    def __init__(self, node):
        Result.__init__(self, node)

    @classmethod
    def _values(cls):
        return IdentifiableResult

class FeatureTraits(IdentifiableTraits, RenderableTraits):

    def closest_distance(self, a0: StringTraits, a1: FloatTraits, a2: QueryTraits) -> FloatTraits:
        return closest_distance(self, a0, a1, a2)

    def has_key(self, a0: StringTraits) -> BoolTraits:
        return has_key(self, a0)

    def paths_to_reach(self, a0: StringTraits, a1: FloatTraits, a2: QueryTraits) -> FeatureIDIntCollectionResult:
        return paths_to_reach(self, a0, a1, a2)

    def reachable(self, a0: StringTraits, a1: FloatTraits, a2: QueryTraits) -> FeatureCollectionResult:
        return reachable(self, a0, a1, a2)

    def point_features(self) -> PointFeatureCollectionResult:
        return point_features(self)

    def tile_ids(self) -> FeatureIDIntCollectionResult:
        return tile_ids(self)

    def reachable_points(self, a0: StringTraits, a1: FloatTraits, a2: QueryTraits) -> PointFeatureCollectionResult:
        return reachable_points(self, a0, a1, a2)

    def closest(self, a0: StringTraits, a1: FloatTraits, a2: QueryTraits) -> FeatureTraits:
        return closest(self, a0, a1, a2)

    def tile_ids_hex(self) -> FeatureIDStringCollectionResult:
        return tile_ids_hex(self)

    def reachable_area(self, a0: StringTraits, a1: FloatTraits) -> FloatTraits:
        return reachable_area(self, a0, a1)

    def connect_to_network(self) -> ChangeTraits:
        return connect_to_network(self)

    @classmethod
    def _collection(cls):
        return FeatureCollectionResult

class FeatureValuesTraits(IdentifiableValuesTraits, RenderableValuesTraits):

    def closest_distance(self, a0: StringTraits, a1: FloatTraits, a2: QueryTraits) -> AnyFloatCollectionTraits:
        return self.map(Lambda(lambda x: closest_distance(x, a0, a1, a2), [self._values()]))

    def has_key(self, a0: StringTraits) -> CollectionTraits:
        return self.map(Lambda(lambda x: has_key(x, a0), [self._values()]))

    def paths_to_reach(self, a0: StringTraits, a1: FloatTraits, a2: QueryTraits) -> CollectionTraits:
        return self.map(Lambda(lambda x: paths_to_reach(x, a0, a1, a2), [self._values()]))

    def reachable(self, a0: StringTraits, a1: FloatTraits, a2: QueryTraits) -> CollectionTraits:
        return self.map(Lambda(lambda x: reachable(x, a0, a1, a2), [self._values()]))

    def point_features(self) -> CollectionTraits:
        return self.map(Lambda(point_features, [self._values()]))

    def tile_ids(self) -> CollectionTraits:
        return self.map(Lambda(tile_ids, [self._values()]))

    def reachable_points(self, a0: StringTraits, a1: FloatTraits, a2: QueryTraits) -> CollectionTraits:
        return self.map(Lambda(lambda x: reachable_points(x, a0, a1, a2), [self._values()]))

    def closest(self, a0: StringTraits, a1: FloatTraits, a2: QueryTraits) -> FeatureCollectionTraits:
        return self.map(Lambda(lambda x: closest(x, a0, a1, a2), [self._values()]))

    def tile_ids_hex(self) -> CollectionTraits:
        return self.map(Lambda(tile_ids_hex, [self._values()]))

    def reachable_area(self, a0: StringTraits, a1: FloatTraits) -> AnyFloatCollectionTraits:
        return self.map(Lambda(lambda x: reachable_area(x, a0, a1), [self._values()]))

    def connect_to_network(self) -> CollectionTraits:
        return self.map(Lambda(connect_to_network, [self._values()]))

class AnyFeatureValuesResult(Result, FeatureValuesTraits, CollectionTraits):

    def __init__(self, node):
        Result.__init__(self, node)

    @classmethod
    def _values(cls):
        return FeatureResult

class NumberTraits:

    def divide(self, a0: NumberTraits) -> NumberTraits:
        return divide(self, a0)

    @classmethod
    def _collection(cls):
        return CollectionResult

class NumberValuesTraits:

    def divide(self, a0: NumberTraits) -> CollectionTraits:
        return self.map(Lambda(lambda x: divide(x, a0), [self._values()]))

class AnyNumberValuesResult(Result, NumberValuesTraits, CollectionTraits):

    def __init__(self, node):
        Result.__init__(self, node)

    @classmethod
    def _values(cls):
        return NumberResult

class AreaTraits(GeometryTraits):

    def area(self) -> FloatTraits:
        return area(self)

    def s2_points(self, a0: IntTraits, a1: IntTraits) -> StringPointCollectionResult:
        return s2_points(self, a0, a1)

    def snap_area_edges(self, a0: QueryTraits, a1: FloatTraits) -> AreaTraits:
        return snap_area_edges(self, a0, a1)

    def within(self) -> QueryTraits:
        return within(self)

    def s2_covering(self, a0: IntTraits, a1: IntTraits) -> StringStringCollectionResult:
        return s2_covering(self, a0, a1)

    def s2_grid(self, a0: IntTraits) -> StringStringCollectionResult:
        return s2_grid(self, a0)

    @classmethod
    def _collection(cls):
        return AreaCollectionResult

class AreaValuesTraits(GeometryValuesTraits):

    def area(self) -> AnyFloatCollectionTraits:
        return self.map(Lambda(area, [self._values()]))

    def s2_points(self, a0: IntTraits, a1: IntTraits) -> CollectionTraits:
        return self.map(Lambda(lambda x: s2_points(x, a0, a1), [self._values()]))

    def snap_area_edges(self, a0: QueryTraits, a1: FloatTraits) -> AreaCollectionTraits:
        return self.map(Lambda(lambda x: snap_area_edges(x, a0, a1), [self._values()]))

    def within(self) -> CollectionTraits:
        return self.map(Lambda(within, [self._values()]))

    def s2_covering(self, a0: IntTraits, a1: IntTraits) -> CollectionTraits:
        return self.map(Lambda(lambda x: s2_covering(x, a0, a1), [self._values()]))

    def s2_grid(self, a0: IntTraits) -> CollectionTraits:
        return self.map(Lambda(lambda x: s2_grid(x, a0), [self._values()]))

class AnyAreaValuesResult(Result, AreaValuesTraits, CollectionTraits):

    def __init__(self, node):
        Result.__init__(self, node)

    @classmethod
    def _values(cls):
        return AreaResult

class PointTraits(GeometryTraits):

    def cap_polygon(self, a0: FloatTraits) -> AreaTraits:
        return cap_polygon(self, a0)

    def distance_meters(self, a0: PointTraits) -> FloatTraits:
        return distance_meters(self, a0)

    def intersecting_cap(self, a0: FloatTraits) -> QueryTraits:
        return intersecting_cap(self, a0)

    def add_point(self, a0: StringPointCollectionResult) -> StringPointCollectionResult:
        return add_point(self, a0)

    def within_cap(self, a0: FloatTraits) -> QueryTraits:
        return within_cap(self, a0)

    def rectangle_polygon(self, a0: PointTraits) -> AreaTraits:
        return rectangle_polygon(self, a0)

    def sightline(self, a0: FloatTraits) -> AreaTraits:
        return sightline(self, a0)

    @classmethod
    def _collection(cls):
        return PointCollectionResult

class PointValuesTraits(GeometryValuesTraits):

    def cap_polygon(self, a0: FloatTraits) -> AreaCollectionTraits:
        return self.map(Lambda(lambda x: cap_polygon(x, a0), [self._values()]))

    def distance_meters(self, a0: PointTraits) -> AnyFloatCollectionTraits:
        return self.map(Lambda(lambda x: distance_meters(x, a0), [self._values()]))

    def intersecting_cap(self, a0: FloatTraits) -> CollectionTraits:
        return self.map(Lambda(lambda x: intersecting_cap(x, a0), [self._values()]))

    def add_point(self, a0: StringPointCollectionResult) -> CollectionTraits:
        return self.map(Lambda(lambda x: add_point(x, a0), [self._values()]))

    def within_cap(self, a0: FloatTraits) -> CollectionTraits:
        return self.map(Lambda(lambda x: within_cap(x, a0), [self._values()]))

    def rectangle_polygon(self, a0: PointTraits) -> AreaCollectionTraits:
        return self.map(Lambda(lambda x: rectangle_polygon(x, a0), [self._values()]))

    def sightline(self, a0: FloatTraits) -> AreaCollectionTraits:
        return self.map(Lambda(lambda x: sightline(x, a0), [self._values()]))

class AnyPointValuesResult(Result, PointValuesTraits, CollectionTraits):

    def __init__(self, node):
        Result.__init__(self, node)

    @classmethod
    def _values(cls):
        return PointResult

class StringTraits:

    def s2_polygon(self) -> AreaTraits:
        return s2_polygon(self)

    def tagged(self, a0: StringTraits) -> QueryTraits:
        return tagged(self, a0)

    def debug_all_query(self) -> QueryTraits:
        return debug_all_query(self)

    def keyed(self) -> QueryTraits:
        return keyed(self)

    def tag(self, a0: StringTraits) -> TagTraits:
        return tag(self, a0)

    def parse_geojson(self) -> GeoJSONTraits:
        return parse_geojson(self)

    def import_geojson_file(self, a0: StringTraits) -> ChangeTraits:
        return import_geojson_file(self, a0)

    def s2_center(self) -> PointTraits:
        return s2_center(self)

    def export_world(self) -> IntTraits:
        return export_world(self)

    @classmethod
    def _collection(cls):
        return CollectionResult

class StringValuesTraits:

    def s2_polygon(self) -> AreaCollectionTraits:
        return self.map(Lambda(s2_polygon, [self._values()]))

    def tagged(self, a0: StringTraits) -> CollectionTraits:
        return self.map(Lambda(lambda x: tagged(x, a0), [self._values()]))

    def debug_all_query(self) -> CollectionTraits:
        return self.map(Lambda(debug_all_query, [self._values()]))

    def keyed(self) -> CollectionTraits:
        return self.map(Lambda(keyed, [self._values()]))

    def tag(self, a0: StringTraits) -> CollectionTraits:
        return self.map(Lambda(lambda x: tag(x, a0), [self._values()]))

    def parse_geojson(self) -> CollectionTraits:
        return self.map(Lambda(parse_geojson, [self._values()]))

    def import_geojson_file(self, a0: StringTraits) -> CollectionTraits:
        return self.map(Lambda(lambda x: import_geojson_file(x, a0), [self._values()]))

    def s2_center(self) -> PointCollectionTraits:
        return self.map(Lambda(s2_center, [self._values()]))

    def export_world(self) -> CollectionTraits:
        return self.map(Lambda(export_world, [self._values()]))

class AnyStringValuesResult(Result, StringValuesTraits, CollectionTraits):

    def __init__(self, node):
        Result.__init__(self, node)

    @classmethod
    def _values(cls):
        return StringResult

class PathTraits(GeometryTraits):

    def ordered_join(self, a0: PathTraits) -> PathTraits:
        return ordered_join(self, a0)

    def join(self, a0: PathTraits) -> PathTraits:
        return join(self, a0)

    def sample_points(self, a0: FloatTraits) -> StringPointCollectionResult:
        return sample_points(self, a0)

    def distance_to_point_meters(self, a0: PointTraits) -> FloatTraits:
        return distance_to_point_meters(self, a0)

    def interpolate(self, a0: FloatTraits) -> PointTraits:
        return interpolate(self, a0)

    @classmethod
    def _collection(cls):
        return PathCollectionResult

class PathValuesTraits(GeometryValuesTraits):

    def ordered_join(self, a0: PathTraits) -> PathCollectionTraits:
        return self.map(Lambda(lambda x: ordered_join(x, a0), [self._values()]))

    def join(self, a0: PathTraits) -> PathCollectionTraits:
        return self.map(Lambda(lambda x: join(x, a0), [self._values()]))

    def sample_points(self, a0: FloatTraits) -> CollectionTraits:
        return self.map(Lambda(lambda x: sample_points(x, a0), [self._values()]))

    def distance_to_point_meters(self, a0: PointTraits) -> AnyFloatCollectionTraits:
        return self.map(Lambda(lambda x: distance_to_point_meters(x, a0), [self._values()]))

    def interpolate(self, a0: FloatTraits) -> PointCollectionTraits:
        return self.map(Lambda(lambda x: interpolate(x, a0), [self._values()]))

class AnyPathValuesResult(Result, PathValuesTraits, CollectionTraits):

    def __init__(self, node):
        Result.__init__(self, node)

    @classmethod
    def _values(cls):
        return PathResult

class TagTraits:

    def float_value(self) -> FloatTraits:
        return float_value(self)

    def value(self) -> StringTraits:
        return value(self)

    def int_value(self) -> IntTraits:
        return int_value(self)

    @classmethod
    def _collection(cls):
        return CollectionResult

class TagValuesTraits:

    def float_value(self) -> AnyFloatCollectionTraits:
        return self.map(Lambda(float_value, [self._values()]))

    def value(self) -> CollectionTraits:
        return self.map(Lambda(value, [self._values()]))

    def int_value(self) -> CollectionTraits:
        return self.map(Lambda(int_value, [self._values()]))

class AnyTagValuesResult(Result, TagValuesTraits, CollectionTraits):

    def __init__(self, node):
        Result.__init__(self, node)

    @classmethod
    def _values(cls):
        return TagResult

class IdentifiablePointTraits:

    def point_paths(self) -> PathFeatureCollectionResult:
        return point_paths(self)

    @classmethod
    def _collection(cls):
        return CollectionResult

class IdentifiablePointValuesTraits:

    def point_paths(self) -> CollectionTraits:
        return self.map(Lambda(point_paths, [self._values()]))

class AnyIdentifiablePointValuesResult(Result, IdentifiablePointValuesTraits, CollectionTraits):

    def __init__(self, node):
        Result.__init__(self, node)

    @classmethod
    def _values(cls):
        return IdentifiablePointResult

class AreaFeatureTraits(AreaTraits, FeatureTraits):
    @classmethod
    def _collection(cls):
        return AreaFeatureCollectionResult

class AreaFeatureValuesTraits(AreaValuesTraits, FeatureValuesTraits):
    pass

class AnyAreaFeatureValuesResult(Result, AreaFeatureValuesTraits, CollectionTraits):

    def __init__(self, node):
        Result.__init__(self, node)

    @classmethod
    def _values(cls):
        return AreaFeatureResult

class AnyTraits:

    def pair(self, a0: AnyTraits) -> PairTraits:
        return pair(self, a0)

    def gt(self, a0: AnyTraits) -> BoolTraits:
        return gt(self, a0)

    def collection(self, a0: AnyTraits) -> CollectionResult:
        return collection(self, a0)

    @classmethod
    def _collection(cls):
        return CollectionResult

class AnyValuesTraits:

    def pair(self, a0: AnyTraits) -> CollectionTraits:
        return self.map(Lambda(lambda x: pair(x, a0), [self._values()]))

    def gt(self, a0: AnyTraits) -> CollectionTraits:
        return self.map(Lambda(lambda x: gt(x, a0), [self._values()]))

    def collection(self, a0: AnyTraits) -> CollectionTraits:
        return self.map(Lambda(lambda x: collection(x, a0), [self._values()]))

class AnyAnyValuesResult(Result, AnyValuesTraits, CollectionTraits):

    def __init__(self, node):
        Result.__init__(self, node)

    @classmethod
    def _values(cls):
        return AnyResult

class PointFeatureTraits(PointTraits, IdentifiablePointTraits, FeatureTraits):

    def degree(self) -> IntTraits:
        return degree(self)

    def connect(self, a0: PointFeatureTraits) -> ChangeTraits:
        return connect(self, a0)

    @classmethod
    def _collection(cls):
        return PointFeatureCollectionResult

class PointFeatureValuesTraits(PointValuesTraits, IdentifiablePointValuesTraits, FeatureValuesTraits):

    def degree(self) -> CollectionTraits:
        return self.map(Lambda(degree, [self._values()]))

    def connect(self, a0: PointFeatureTraits) -> CollectionTraits:
        return self.map(Lambda(lambda x: connect(x, a0), [self._values()]))

class AnyPointFeatureValuesResult(Result, PointFeatureValuesTraits, CollectionTraits):

    def __init__(self, node):
        Result.__init__(self, node)

    @classmethod
    def _values(cls):
        return PointFeatureResult

class RelationFeatureTraits(FeatureTraits, GeometryTraits):
    @classmethod
    def _collection(cls):
        return RelationFeatureCollectionResult

class RelationFeatureValuesTraits(FeatureValuesTraits, GeometryValuesTraits):
    pass

class AnyRelationFeatureValuesResult(Result, RelationFeatureValuesTraits, CollectionTraits):

    def __init__(self, node):
        Result.__init__(self, node)

    @classmethod
    def _values(cls):
        return RelationFeatureResult

class FloatTraits(NumberTraits):

    def ll(self, a0: FloatTraits) -> PointTraits:
        return ll(self, a0)

    @classmethod
    def _collection(cls):
        return AnyFloatCollectionResult

class FloatValuesTraits(NumberValuesTraits):

    def ll(self, a0: FloatTraits) -> PointCollectionTraits:
        return self.map(Lambda(lambda x: ll(x, a0), [self._values()]))

class AnyFloatValuesResult(Result, FloatValuesTraits, CollectionTraits):

    def __init__(self, node):
        Result.__init__(self, node)

    @classmethod
    def _values(cls):
        return FloatResult

class PointCollectionTraits(CollectionTraits, PointValuesTraits):
    @classmethod
    def _collection(cls):
        return CollectionResult

class IntTraits(NumberTraits):

    def clamp(self, a0: IntTraits, a1: IntTraits) -> IntTraits:
        return clamp(self, a0, a1)

    def add_ints(self, a0: IntTraits) -> IntTraits:
        return add_ints(self, a0)

    def divide_int(self, a0: FloatTraits) -> FloatTraits:
        return divide_int(self, a0)

    @classmethod
    def _collection(cls):
        return CollectionResult

class IntValuesTraits(NumberValuesTraits):

    def clamp(self, a0: IntTraits, a1: IntTraits) -> CollectionTraits:
        return self.map(Lambda(lambda x: clamp(x, a0, a1), [self._values()]))

    def add_ints(self, a0: IntTraits) -> CollectionTraits:
        return self.map(Lambda(lambda x: add_ints(x, a0), [self._values()]))

    def divide_int(self, a0: FloatTraits) -> AnyFloatCollectionTraits:
        return self.map(Lambda(lambda x: divide_int(x, a0), [self._values()]))

class AnyIntValuesResult(Result, IntValuesTraits, CollectionTraits):

    def __init__(self, node):
        Result.__init__(self, node)

    @classmethod
    def _values(cls):
        return IntResult

class PathFeatureTraits(PathTraits, FeatureTraits):

    def length(self) -> FloatTraits:
        return length(self)

    @classmethod
    def _collection(cls):
        return PathFeatureCollectionResult

class PathFeatureValuesTraits(PathValuesTraits, FeatureValuesTraits):

    def length(self) -> AnyFloatCollectionTraits:
        return self.map(Lambda(length, [self._values()]))

class AnyPathFeatureValuesResult(Result, PathFeatureValuesTraits, CollectionTraits):

    def __init__(self, node):
        Result.__init__(self, node)

    @classmethod
    def _values(cls):
        return PathFeatureResult

class FeatureIDTraits(IdentifiableTraits):

    def find_feature(self) -> FeatureTraits:
        return find_feature(self)

    def find_point(self) -> PointFeatureTraits:
        return find_point(self)

    def find_area(self) -> AreaFeatureTraits:
        return find_area(self)

    def find_path(self) -> PathFeatureTraits:
        return find_path(self)

    def find_relation(self) -> RelationFeatureTraits:
        return find_relation(self)

    @classmethod
    def _collection(cls):
        return CollectionResult

class FeatureIDValuesTraits(IdentifiableValuesTraits):

    def find_feature(self) -> FeatureCollectionTraits:
        return self.map(Lambda(find_feature, [self._values()]))

    def find_point(self) -> PointFeatureCollectionTraits:
        return self.map(Lambda(find_point, [self._values()]))

    def find_area(self) -> AreaFeatureCollectionTraits:
        return self.map(Lambda(find_area, [self._values()]))

    def find_path(self) -> PathFeatureCollectionTraits:
        return self.map(Lambda(find_path, [self._values()]))

    def find_relation(self) -> RelationFeatureCollectionTraits:
        return self.map(Lambda(find_relation, [self._values()]))

class AnyFeatureIDValuesResult(Result, FeatureIDValuesTraits, CollectionTraits):

    def __init__(self, node):
        Result.__init__(self, node)

    @classmethod
    def _values(cls):
        return FeatureIDResult

class AreaCollectionTraits(CollectionTraits, AreaValuesTraits):
    @classmethod
    def _collection(cls):
        return CollectionResult

class PathCollectionTraits(CollectionTraits, PathValuesTraits):

    def sample_points_along_paths(self, a0: FloatTraits) -> PointCollectionResult:
        return sample_points_along_paths(self, a0)

    @classmethod
    def _collection(cls):
        return CollectionResult

class AnyRenderableCollectionTraits(CollectionTraits, RenderableValuesTraits):

    def to_geojson_collection(self) -> GeoJSONTraits:
        return to_geojson_collection(self)

    @classmethod
    def _collection(cls):
        return CollectionResult

class QueryProtoTraits:
    @classmethod
    def _collection(cls):
        return CollectionResult

class StringPointCollectionTraits(CollectionTraits, PointValuesTraits):
    @classmethod
    def _collection(cls):
        return CollectionResult

class IntStringCollectionTraits(CollectionTraits, StringValuesTraits):
    @classmethod
    def _collection(cls):
        return CollectionResult

class RelationFeatureCollectionTraits(CollectionTraits, RelationFeatureValuesTraits):
    @classmethod
    def _collection(cls):
        return CollectionResult

class AnyFloatCollectionTraits(CollectionTraits, FloatValuesTraits):

    def percentiles(self) -> AnyFloatCollectionResult:
        return percentiles(self)

    @classmethod
    def _collection(cls):
        return CollectionResult

class PointFeatureCollectionTraits(PointCollectionTraits, PointFeatureValuesTraits):

    def containing_areas(self, a0: QueryTraits) -> AreaFeatureCollectionResult:
        return containing_areas(self, a0)

    @classmethod
    def _collection(cls):
        return CollectionResult

class StringStringCollectionTraits(CollectionTraits, StringValuesTraits):
    @classmethod
    def _collection(cls):
        return CollectionResult

class AreaFeatureCollectionTraits(AreaCollectionTraits, AreaFeatureValuesTraits):
    @classmethod
    def _collection(cls):
        return CollectionResult

class StringAreaCollectionTraits(CollectionTraits, AreaValuesTraits):
    @classmethod
    def _collection(cls):
        return CollectionResult

class FeatureCollectionTraits(CollectionTraits, FeatureValuesTraits):

    def building_access(self, a0: FloatTraits, a1: StringTraits) -> FeatureIDFeatureIDCollectionResult:
        return building_access(self, a0, a1)

    @classmethod
    def _collection(cls):
        return CollectionResult

class IntTagCollectionTraits(CollectionTraits, TagValuesTraits):
    @classmethod
    def _collection(cls):
        return CollectionResult

class GeoJSONTraits:

    def geojson_areas(self) -> StringAreaCollectionResult:
        return geojson_areas(self)

    def map_geometries(self, a0: Callable[[GeometryTraits],GeometryTraits]) -> GeoJSONTraits:
        return map_geometries(self, a0)

    def import_geojson(self, a0: StringTraits) -> ChangeTraits:
        return import_geojson(self, a0)

    @classmethod
    def _collection(cls):
        return CollectionResult

class FeatureIDIntCollectionTraits(CollectionTraits, IntValuesTraits):
    @classmethod
    def _collection(cls):
        return CollectionResult

class QueryTraits:

    def and_(self, a0: QueryTraits) -> QueryTraits:
        return and_(self, a0)

    def find_areas(self) -> AreaFeatureCollectionResult:
        return find_areas(self)

    def find_points(self) -> PointFeatureCollectionResult:
        return find_points(self)

    def or_(self, a0: QueryTraits) -> QueryTraits:
        return or_(self, a0)

    def find_relations(self) -> RelationFeatureCollectionResult:
        return find_relations(self)

    def find_paths(self) -> PathFeatureCollectionResult:
        return find_paths(self)

    def find(self) -> FeatureCollectionResult:
        return find(self)

    @classmethod
    def _collection(cls):
        return CollectionResult

class BoolTraits:
    @classmethod
    def _collection(cls):
        return CollectionResult

class PairTraits:

    def second(self) -> AnyTraits:
        return second(self)

    def first(self) -> AnyTraits:
        return first(self)

    @classmethod
    def _collection(cls):
        return CollectionResult

class PathFeatureCollectionTraits(PathCollectionTraits, PathFeatureValuesTraits):
    @classmethod
    def _collection(cls):
        return CollectionResult

class FeatureIDTagCollectionTraits(CollectionTraits, TagValuesTraits):

    def add_tags(self) -> ChangeTraits:
        return add_tags(self)

    @classmethod
    def _collection(cls):
        return CollectionResult

class FeatureIDStringCollectionTraits(CollectionTraits, StringValuesTraits):

    def remove_tags(self) -> ChangeTraits:
        return remove_tags(self)

    @classmethod
    def _collection(cls):
        return CollectionResult

class FeatureIDFeatureIDCollectionTraits(CollectionTraits, FeatureIDValuesTraits):
    @classmethod
    def _collection(cls):
        return CollectionResult

class ChangeTraits:

    def with_change(self, a0: Callable[[],AnyTraits]) -> AnyTraits:
        return with_change(self, a0)

    @classmethod
    def _collection(cls):
        return CollectionResult

class CollectionResult(Result, CollectionTraits):
    def __init__(self, node):
        Result.__init__(self, node)

    @classmethod
    def _values(cls):
        return AnyResult

class RenderableResult(Result, RenderableTraits):
    def __init__(self, node):
        Result.__init__(self, node)

class GeometryResult(Result, GeometryTraits):
    def __init__(self, node):
        Result.__init__(self, node)

class IdentifiableResult(Result, IdentifiableTraits):
    def __init__(self, node):
        Result.__init__(self, node)

class FeatureResult(Result, FeatureTraits):
    def __init__(self, node):
        Result.__init__(self, node)

class NumberResult(Result, NumberTraits):
    def __init__(self, node):
        Result.__init__(self, node)

class AreaResult(Result, AreaTraits):
    def __init__(self, node):
        Result.__init__(self, node)

class PointResult(Result, PointTraits):
    def __init__(self, node):
        Result.__init__(self, node)

class StringResult(Result, StringTraits):
    def __init__(self, node):
        Result.__init__(self, node)

class PathResult(Result, PathTraits):
    def __init__(self, node):
        Result.__init__(self, node)

class TagResult(Result, TagTraits):
    def __init__(self, node):
        Result.__init__(self, node)

class IdentifiablePointResult(Result, IdentifiablePointTraits):
    def __init__(self, node):
        Result.__init__(self, node)

class AreaFeatureResult(Result, AreaFeatureTraits):
    def __init__(self, node):
        Result.__init__(self, node)

class AnyResult(Result, AnyTraits):
    def __init__(self, node):
        Result.__init__(self, node)

class PointFeatureResult(Result, PointFeatureTraits):
    def __init__(self, node):
        Result.__init__(self, node)

class RelationFeatureResult(Result, RelationFeatureTraits):
    def __init__(self, node):
        Result.__init__(self, node)

class FloatResult(Result, FloatTraits):
    def __init__(self, node):
        Result.__init__(self, node)

class PointCollectionResult(Result, PointCollectionTraits):
    def __init__(self, node):
        Result.__init__(self, node)

    @classmethod
    def _values(cls):
        return PointResult

class IntResult(Result, IntTraits):
    def __init__(self, node):
        Result.__init__(self, node)

class PathFeatureResult(Result, PathFeatureTraits):
    def __init__(self, node):
        Result.__init__(self, node)

class FeatureIDResult(Result, FeatureIDTraits):
    def __init__(self, node):
        Result.__init__(self, node)

class AreaCollectionResult(Result, AreaCollectionTraits):
    def __init__(self, node):
        Result.__init__(self, node)

    @classmethod
    def _values(cls):
        return AreaResult

class PathCollectionResult(Result, PathCollectionTraits):
    def __init__(self, node):
        Result.__init__(self, node)

    @classmethod
    def _values(cls):
        return PathResult

class AnyRenderableCollectionResult(Result, AnyRenderableCollectionTraits):
    def __init__(self, node):
        Result.__init__(self, node)

    @classmethod
    def _values(cls):
        return RenderableResult

class QueryProtoResult(Result, QueryProtoTraits):
    def __init__(self, node):
        Result.__init__(self, node)

class StringPointCollectionResult(Result, StringPointCollectionTraits):
    def __init__(self, node):
        Result.__init__(self, node)

    @classmethod
    def _values(cls):
        return PointResult

class IntStringCollectionResult(Result, IntStringCollectionTraits):
    def __init__(self, node):
        Result.__init__(self, node)

    @classmethod
    def _values(cls):
        return StringResult

class RelationFeatureCollectionResult(Result, RelationFeatureCollectionTraits):
    def __init__(self, node):
        Result.__init__(self, node)

    @classmethod
    def _values(cls):
        return RelationFeatureResult

class AnyFloatCollectionResult(Result, AnyFloatCollectionTraits):
    def __init__(self, node):
        Result.__init__(self, node)

    @classmethod
    def _values(cls):
        return FloatResult

class PointFeatureCollectionResult(Result, PointFeatureCollectionTraits):
    def __init__(self, node):
        Result.__init__(self, node)

    @classmethod
    def _values(cls):
        return PointFeatureResult

class StringStringCollectionResult(Result, StringStringCollectionTraits):
    def __init__(self, node):
        Result.__init__(self, node)

    @classmethod
    def _values(cls):
        return StringResult

class AreaFeatureCollectionResult(Result, AreaFeatureCollectionTraits):
    def __init__(self, node):
        Result.__init__(self, node)

    @classmethod
    def _values(cls):
        return AreaFeatureResult

class StringAreaCollectionResult(Result, StringAreaCollectionTraits):
    def __init__(self, node):
        Result.__init__(self, node)

    @classmethod
    def _values(cls):
        return AreaResult

class FeatureCollectionResult(Result, FeatureCollectionTraits):
    def __init__(self, node):
        Result.__init__(self, node)

    @classmethod
    def _values(cls):
        return FeatureResult

class IntTagCollectionResult(Result, IntTagCollectionTraits):
    def __init__(self, node):
        Result.__init__(self, node)

    @classmethod
    def _values(cls):
        return TagResult

class GeoJSONResult(Result, GeoJSONTraits):
    def __init__(self, node):
        Result.__init__(self, node)

class FeatureIDIntCollectionResult(Result, FeatureIDIntCollectionTraits):
    def __init__(self, node):
        Result.__init__(self, node)

    @classmethod
    def _values(cls):
        return IntResult

class QueryResult(Result, QueryTraits):
    def __init__(self, node):
        Result.__init__(self, node)

class BoolResult(Result, BoolTraits):
    def __init__(self, node):
        Result.__init__(self, node)

class PairResult(Result, PairTraits):
    def __init__(self, node):
        Result.__init__(self, node)

class PathFeatureCollectionResult(Result, PathFeatureCollectionTraits):
    def __init__(self, node):
        Result.__init__(self, node)

    @classmethod
    def _values(cls):
        return PathFeatureResult

class FeatureIDTagCollectionResult(Result, FeatureIDTagCollectionTraits):
    def __init__(self, node):
        Result.__init__(self, node)

    @classmethod
    def _values(cls):
        return TagResult

class FeatureIDStringCollectionResult(Result, FeatureIDStringCollectionTraits):
    def __init__(self, node):
        Result.__init__(self, node)

    @classmethod
    def _values(cls):
        return StringResult

class FeatureIDFeatureIDCollectionResult(Result, FeatureIDFeatureIDCollectionTraits):
    def __init__(self, node):
        Result.__init__(self, node)

    @classmethod
    def _values(cls):
        return FeatureIDResult

class ChangeResult(Result, ChangeTraits):
    def __init__(self, node):
        Result.__init__(self, node)

class FunctionAreaGeometryResult(Result, Callable[[AreaTraits],GeometryTraits]):

    def __init__(self, node):
        Result.__init__(self, node)

    def __call__(self, a0 : AreaTraits) -> GeometryTraits:
        raise NotImplementedError()

class FunctionAnyAnyResult(Result, Callable[[AnyTraits],AnyTraits]):

    def __init__(self, node):
        Result.__init__(self, node)

    def __call__(self, a0 : AnyTraits) -> AnyTraits:
        raise NotImplementedError()

class FunctionPointGeometryResult(Result, Callable[[PointTraits],GeometryTraits]):

    def __init__(self, node):
        Result.__init__(self, node)

    def __call__(self, a0 : PointTraits) -> GeometryTraits:
        raise NotImplementedError()

class FunctionPairAnyResult(Result, Callable[[PairTraits],AnyTraits]):

    def __init__(self, node):
        Result.__init__(self, node)

    def __call__(self, a0 : PairTraits) -> AnyTraits:
        raise NotImplementedError()

class FunctionPathGeometryResult(Result, Callable[[PathTraits],GeometryTraits]):

    def __init__(self, node):
        Result.__init__(self, node)

    def __call__(self, a0 : PathTraits) -> GeometryTraits:
        raise NotImplementedError()

class FunctionAnyBoolResult(Result, Callable[[AnyTraits],BoolTraits]):

    def __init__(self, node):
        Result.__init__(self, node)

    def __call__(self, a0 : AnyTraits) -> BoolTraits:
        raise NotImplementedError()

class FunctionAnyResult(Result, Callable[[],AnyTraits]):

    def __init__(self, node):
        Result.__init__(self, node)

    def __call__(self, ) -> AnyTraits:
        raise NotImplementedError()

class FunctionGeometryGeometryResult(Result, Callable[[GeometryTraits],GeometryTraits]):

    def __init__(self, node):
        Result.__init__(self, node)

    def __call__(self, a0 : GeometryTraits) -> GeometryTraits:
        raise NotImplementedError()

def closest_distance(a0: FeatureTraits, a1: StringTraits, a2: FloatTraits, a3: QueryTraits) -> FloatTraits:
    return FloatResult(Call(Symbol('closest-distance'), [a0, a1, a2, a3]))

def cap_polygon(a0: PointTraits, a1: FloatTraits) -> AreaTraits:
    return AreaResult(Call(Symbol('cap-polygon'), [a0, a1]))

def apply_to_area(a0: Callable[[AreaTraits],GeometryTraits]) -> Callable[[GeometryTraits],GeometryTraits]:
    return FunctionGeometryGeometryResult(Call(Symbol('apply-to-area'), [a0]))

def pair(a0: AnyTraits, a1: AnyTraits) -> PairTraits:
    return PairResult(Call(Symbol('pair'), [a0, a1]))

def and_(a0: QueryTraits, a1: QueryTraits) -> QueryTraits:
    return QueryResult(Call(Symbol('and'), [a0, a1]))

def take(a0: CollectionResult, a1: IntTraits) -> CollectionResult:
    return CollectionResult(Call(Symbol('take'), [a0, a1]))

def find_feature(a0: FeatureIDTraits) -> FeatureTraits:
    return FeatureResult(Call(Symbol('find-feature'), [a0]))

def point_paths(a0: IdentifiablePointTraits) -> PathFeatureCollectionResult:
    return PathFeatureCollectionResult(Call(Symbol('point-paths'), [a0]))

def ordered_join(a0: PathTraits, a1: PathTraits) -> PathTraits:
    return PathResult(Call(Symbol('ordered-join'), [a0, a1]))

def area(a0: AreaTraits) -> FloatTraits:
    return FloatResult(Call(Symbol('area'), [a0]))

def find_areas(a0: QueryTraits) -> AreaFeatureCollectionResult:
    return AreaFeatureCollectionResult(Call(Symbol('find-areas'), [a0]))

def degree(a0: PointFeatureTraits) -> IntTraits:
    return IntResult(Call(Symbol('degree'), [a0]))

def s2_polygon(a0: StringTraits) -> AreaTraits:
    return AreaResult(Call(Symbol('s2-polygon'), [a0]))

def has_key(a0: FeatureTraits, a1: StringTraits) -> BoolTraits:
    return BoolResult(Call(Symbol('has-key'), [a0, a1]))

def join(a0: PathTraits, a1: PathTraits) -> PathTraits:
    return PathResult(Call(Symbol('join'), [a0, a1]))

def s2_points(a0: AreaTraits, a1: IntTraits, a2: IntTraits) -> StringPointCollectionResult:
    return StringPointCollectionResult(Call(Symbol('s2-points'), [a0, a1, a2]))

def snap_area_edges(a0: AreaTraits, a1: QueryTraits, a2: FloatTraits) -> AreaTraits:
    return AreaResult(Call(Symbol('snap-area-edges'), [a0, a1, a2]))

def paths_to_reach(a0: FeatureTraits, a1: StringTraits, a2: FloatTraits, a3: QueryTraits) -> FeatureIDIntCollectionResult:
    return FeatureIDIntCollectionResult(Call(Symbol('paths-to-reach'), [a0, a1, a2, a3]))

def to_geojson(a0: RenderableTraits) -> GeoJSONTraits:
    return GeoJSONResult(Call(Symbol('to-geojson'), [a0]))

def map_items(a0: CollectionResult, a1: Callable[[PairTraits],AnyTraits]) -> CollectionResult:
    return CollectionResult(Call(Symbol('map-items'), [a0, a1]))

def tagged(a0: StringTraits, a1: StringTraits) -> QueryTraits:
    return QueryResult(Call(Symbol('tagged'), [a0, a1]))

def clamp(a0: IntTraits, a1: IntTraits, a2: IntTraits) -> IntTraits:
    return IntResult(Call(Symbol('clamp'), [a0, a1, a2]))

def connect(a0: PointFeatureTraits, a1: PointFeatureTraits) -> ChangeTraits:
    return ChangeResult(Call(Symbol('connect'), [a0, a1]))

def to_geojson_collection(a0: AnyRenderableCollectionResult) -> GeoJSONTraits:
    return GeoJSONResult(Call(Symbol('to-geojson-collection'), [a0]))

def apply_to_path(a0: Callable[[PathTraits],GeometryTraits]) -> Callable[[GeometryTraits],GeometryTraits]:
    return FunctionGeometryGeometryResult(Call(Symbol('apply-to-path'), [a0]))

def type_point() -> QueryProtoTraits:
    return QueryProtoResult(Call(Symbol('type-point'), []))

def float_value(a0: TagTraits) -> FloatTraits:
    return FloatResult(Call(Symbol('float-value'), [a0]))

def geojson_areas(a0: GeoJSONTraits) -> StringAreaCollectionResult:
    return StringAreaCollectionResult(Call(Symbol('geojson-areas'), [a0]))

def find_point(a0: FeatureIDTraits) -> PointFeatureTraits:
    return PointFeatureResult(Call(Symbol('find-point'), [a0]))

def find_points(a0: QueryTraits) -> PointFeatureCollectionResult:
    return PointFeatureCollectionResult(Call(Symbol('find-points'), [a0]))

def value(a0: TagTraits) -> StringTraits:
    return StringResult(Call(Symbol('value'), [a0]))

def reachable(a0: FeatureTraits, a1: StringTraits, a2: FloatTraits, a3: QueryTraits) -> FeatureCollectionResult:
    return FeatureCollectionResult(Call(Symbol('reachable'), [a0, a1, a2, a3]))

def distance_meters(a0: PointTraits, a1: PointTraits) -> FloatTraits:
    return FloatResult(Call(Symbol('distance-meters'), [a0, a1]))

def add_tags(a0: FeatureIDTagCollectionResult) -> ChangeTraits:
    return ChangeResult(Call(Symbol('add-tags'), [a0]))

def intersecting_cap(a0: PointTraits, a1: FloatTraits) -> QueryTraits:
    return QueryResult(Call(Symbol('intersecting-cap'), [a0, a1]))

def add_point(a0: PointTraits, a1: StringPointCollectionResult) -> StringPointCollectionResult:
    return StringPointCollectionResult(Call(Symbol('add-point'), [a0, a1]))

def find_area(a0: FeatureIDTraits) -> AreaFeatureTraits:
    return AreaFeatureResult(Call(Symbol('find-area'), [a0]))

def gt(a0: AnyTraits, a1: AnyTraits) -> BoolTraits:
    return BoolResult(Call(Symbol('gt'), [a0, a1]))

def collection(a0: AnyTraits, a1: AnyTraits) -> CollectionResult:
    return CollectionResult(Call(Symbol('collection'), [a0, a1]))

def or_(a0: QueryTraits, a1: QueryTraits) -> QueryTraits:
    return QueryResult(Call(Symbol('or'), [a0, a1]))

def int_value(a0: TagTraits) -> IntTraits:
    return IntResult(Call(Symbol('int-value'), [a0]))

def get_string(a0: IdentifiableTraits, a1: StringTraits) -> StringTraits:
    return StringResult(Call(Symbol('get-string'), [a0, a1]))

def empty_points() -> StringPointCollectionResult:
    return StringPointCollectionResult(Call(Symbol('empty-points'), []))

def second(a0: PairTraits) -> AnyTraits:
    return AnyResult(Call(Symbol('second'), [a0]))

def find_relations(a0: QueryTraits) -> RelationFeatureCollectionResult:
    return RelationFeatureCollectionResult(Call(Symbol('find-relations'), [a0]))

def map_geometries(a0: GeoJSONTraits, a1: Callable[[GeometryTraits],GeometryTraits]) -> GeoJSONTraits:
    return GeoJSONResult(Call(Symbol('map-geometries'), [a0, a1]))

def building_access(a0: FeatureCollectionResult, a1: FloatTraits, a2: StringTraits) -> FeatureIDFeatureIDCollectionResult:
    return FeatureIDFeatureIDCollectionResult(Call(Symbol('building-access'), [a0, a1, a2]))

def debug_all_query(a0: StringTraits) -> QueryTraits:
    return QueryResult(Call(Symbol('debug-all-query'), [a0]))

def count_values(a0: CollectionResult) -> CollectionResult:
    return CollectionResult(Call(Symbol('count-values'), [a0]))

def within(a0: AreaTraits) -> QueryTraits:
    return QueryResult(Call(Symbol('within'), [a0]))

def import_geojson(a0: GeoJSONTraits, a1: StringTraits) -> ChangeTraits:
    return ChangeResult(Call(Symbol('import-geojson'), [a0, a1]))

def add_tag(a0: IdentifiableTraits, a1: TagTraits) -> ChangeTraits:
    return ChangeResult(Call(Symbol('add-tag'), [a0, a1]))

def map_parallel(a0: CollectionResult, a1: Callable[[AnyTraits],AnyTraits]) -> CollectionResult:
    return CollectionResult(Call(Symbol('map-parallel'), [a0, a1]))

def sum_by_key(a0: CollectionResult) -> CollectionResult:
    return CollectionResult(Call(Symbol('sum-by-key'), [a0]))

def keyed(a0: StringTraits) -> QueryTraits:
    return QueryResult(Call(Symbol('keyed'), [a0]))

def sample_points(a0: PathTraits, a1: FloatTraits) -> StringPointCollectionResult:
    return StringPointCollectionResult(Call(Symbol('sample-points'), [a0, a1]))

def find_paths(a0: QueryTraits) -> PathFeatureCollectionResult:
    return PathFeatureCollectionResult(Call(Symbol('find-paths'), [a0]))

def type_path() -> QueryProtoTraits:
    return QueryProtoResult(Call(Symbol('type-path'), []))

def type_area() -> QueryProtoTraits:
    return QueryProtoResult(Call(Symbol('type-area'), []))

def within_cap(a0: PointTraits, a1: FloatTraits) -> QueryTraits:
    return QueryResult(Call(Symbol('within-cap'), [a0, a1]))

def s2_covering(a0: AreaTraits, a1: IntTraits, a2: IntTraits) -> StringStringCollectionResult:
    return StringStringCollectionResult(Call(Symbol('s2-covering'), [a0, a1, a2]))

def centroid(a0: GeometryTraits) -> PointTraits:
    return PointResult(Call(Symbol('centroid'), [a0]))

def find(a0: QueryTraits) -> FeatureCollectionResult:
    return FeatureCollectionResult(Call(Symbol('find'), [a0]))

def all() -> QueryTraits:
    return QueryResult(Call(Symbol('all'), []))

def distance_to_point_meters(a0: PathTraits, a1: PointTraits) -> FloatTraits:
    return FloatResult(Call(Symbol('distance-to-point-meters'), [a0, a1]))

def point_features(a0: FeatureTraits) -> PointFeatureCollectionResult:
    return PointFeatureCollectionResult(Call(Symbol('point-features'), [a0]))

def count(a0: CollectionResult) -> IntTraits:
    return IntResult(Call(Symbol('count'), [a0]))

def tile_ids(a0: FeatureTraits) -> FeatureIDIntCollectionResult:
    return FeatureIDIntCollectionResult(Call(Symbol('tile-ids'), [a0]))

def find_path(a0: FeatureIDTraits) -> PathFeatureTraits:
    return PathFeatureResult(Call(Symbol('find-path'), [a0]))

def tag(a0: StringTraits, a1: StringTraits) -> TagTraits:
    return TagResult(Call(Symbol('tag'), [a0, a1]))

def points(a0: GeometryTraits) -> PointCollectionResult:
    return PointCollectionResult(Call(Symbol('points'), [a0]))

def apply_to_point(a0: Callable[[PointTraits],GeometryTraits]) -> Callable[[GeometryTraits],GeometryTraits]:
    return FunctionGeometryGeometryResult(Call(Symbol('apply-to-point'), [a0]))

def get(a0: IdentifiableTraits, a1: StringTraits) -> TagTraits:
    return TagResult(Call(Symbol('get'), [a0, a1]))

def rectangle_polygon(a0: PointTraits, a1: PointTraits) -> AreaTraits:
    return AreaResult(Call(Symbol('rectangle-polygon'), [a0, a1]))

def s2_grid(a0: AreaTraits, a1: IntTraits) -> StringStringCollectionResult:
    return StringStringCollectionResult(Call(Symbol('s2-grid'), [a0, a1]))

def sightline(a0: PointTraits, a1: FloatTraits) -> AreaTraits:
    return AreaResult(Call(Symbol('sightline'), [a0, a1]))

def parse_geojson(a0: StringTraits) -> GeoJSONTraits:
    return GeoJSONResult(Call(Symbol('parse-geojson'), [a0]))

def remove_tags(a0: FeatureIDStringCollectionResult) -> ChangeTraits:
    return ChangeResult(Call(Symbol('remove-tags'), [a0]))

filter = diagonal_b6.expression._filter

def intersecting(a0: GeometryTraits) -> QueryTraits:
    return QueryResult(Call(Symbol('intersecting'), [a0]))

def percentiles(a0: AnyFloatCollectionResult) -> AnyFloatCollectionResult:
    return AnyFloatCollectionResult(Call(Symbol('percentiles'), [a0]))

def tile_paths(a0: GeometryTraits, a1: IntTraits) -> IntStringCollectionResult:
    return IntStringCollectionResult(Call(Symbol('tile-paths'), [a0, a1]))

def all_tags(a0: IdentifiableTraits) -> IntTagCollectionResult:
    return IntTagCollectionResult(Call(Symbol('all-tags'), [a0]))

def reachable_points(a0: FeatureTraits, a1: StringTraits, a2: FloatTraits, a3: QueryTraits) -> PointFeatureCollectionResult:
    return PointFeatureCollectionResult(Call(Symbol('reachable-points'), [a0, a1, a2, a3]))

def closest(a0: FeatureTraits, a1: StringTraits, a2: FloatTraits, a3: QueryTraits) -> FeatureTraits:
    return FeatureResult(Call(Symbol('closest'), [a0, a1, a2, a3]))

def interpolate(a0: PathTraits, a1: FloatTraits) -> PointTraits:
    return PointResult(Call(Symbol('interpolate'), [a0, a1]))

def import_geojson_file(a0: StringTraits, a1: StringTraits) -> ChangeTraits:
    return ChangeResult(Call(Symbol('import-geojson-file'), [a0, a1]))

def debug_tokens(a0: IdentifiableTraits) -> IntStringCollectionResult:
    return IntStringCollectionResult(Call(Symbol('debug-tokens'), [a0]))

def first(a0: PairTraits) -> AnyTraits:
    return AnyResult(Call(Symbol('first'), [a0]))

def sample_points_along_paths(a0: PathCollectionResult, a1: FloatTraits) -> PointCollectionResult:
    return PointCollectionResult(Call(Symbol('sample-points-along-paths'), [a0, a1]))

def divide(a0: NumberTraits, a1: NumberTraits) -> NumberTraits:
    return NumberResult(Call(Symbol('divide'), [a0, a1]))

def ll(a0: FloatTraits, a1: FloatTraits) -> PointTraits:
    return PointResult(Call(Symbol('ll'), [a0, a1]))

def tile_ids_hex(a0: FeatureTraits) -> FeatureIDStringCollectionResult:
    return FeatureIDStringCollectionResult(Call(Symbol('tile-ids-hex'), [a0]))

def find_relation(a0: FeatureIDTraits) -> RelationFeatureTraits:
    return RelationFeatureResult(Call(Symbol('find-relation'), [a0]))

def add_ints(a0: IntTraits, a1: IntTraits) -> IntTraits:
    return IntResult(Call(Symbol('add-ints'), [a0, a1]))

def flattern(a0: CollectionResult) -> CollectionResult:
    return CollectionResult(Call(Symbol('flattern'), [a0]))

def get_int(a0: IdentifiableTraits, a1: StringTraits) -> IntTraits:
    return IntResult(Call(Symbol('get-int'), [a0, a1]))

def s2_center(a0: StringTraits) -> PointTraits:
    return PointResult(Call(Symbol('s2-center'), [a0]))

def with_change(a0: ChangeTraits, a1: Callable[[],AnyTraits]) -> AnyTraits:
    return AnyResult(Call(Symbol('with-change'), [a0, a1]))

def export_world(a0: StringTraits) -> IntTraits:
    return IntResult(Call(Symbol('export-world'), [a0]))

def containing_areas(a0: PointFeatureCollectionResult, a1: QueryTraits) -> AreaFeatureCollectionResult:
    return AreaFeatureCollectionResult(Call(Symbol('containing-areas'), [a0, a1]))

def count_tag_value(a0: IdentifiableTraits, a1: StringTraits) -> CollectionResult:
    return CollectionResult(Call(Symbol('count-tag-value'), [a0, a1]))

def reachable_area(a0: FeatureTraits, a1: StringTraits, a2: FloatTraits) -> FloatTraits:
    return FloatResult(Call(Symbol('reachable-area'), [a0, a1, a2]))

def get_float(a0: IdentifiableTraits, a1: StringTraits) -> FloatTraits:
    return FloatResult(Call(Symbol('get-float'), [a0, a1]))

def length(a0: PathFeatureTraits) -> FloatTraits:
    return FloatResult(Call(Symbol('length'), [a0]))

def remove_tag(a0: IdentifiableTraits, a1: StringTraits) -> ChangeTraits:
    return ChangeResult(Call(Symbol('remove-tag'), [a0, a1]))

def divide_int(a0: IntTraits, a1: FloatTraits) -> FloatTraits:
    return FloatResult(Call(Symbol('divide-int'), [a0, a1]))

map = diagonal_b6.expression._map

def connect_to_network(a0: FeatureTraits) -> ChangeTraits:
    return ChangeResult(Call(Symbol('connect-to-network'), [a0]))




# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: features.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


import diagonal_b6.geometry_pb2 as geometry__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x0e\x66\x65\x61tures.proto\x12\x08\x66\x65\x61tures\x1a\x0egeometry.proto\"&\n\x08TagProto\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t\"W\n\x0e\x46\x65\x61tureIDProto\x12#\n\x04type\x18\x01 \x01(\x0e\x32\x15.features.FeatureType\x12\x11\n\tnamespace\x18\x02 \x01(\t\x12\r\n\x05value\x18\x03 \x01(\x04\"\x80\x01\n\x11PointFeatureProto\x12$\n\x02id\x18\x01 \x01(\x0b\x32\x18.features.FeatureIDProto\x12 \n\x04tags\x18\x02 \x03(\x0b\x32\x12.features.TagProto\x12#\n\x05point\x18\x03 \x01(\x0b\x32\x14.geometry.PointProto\"\x9f\x01\n\x10PathFeatureProto\x12$\n\x02id\x18\x01 \x01(\x0b\x32\x18.features.FeatureIDProto\x12 \n\x04tags\x18\x02 \x03(\x0b\x32\x12.features.TagProto\x12-\n\x08\x66\x65\x61tures\x18\x03 \x03(\x0b\x32\x1b.features.PointFeatureProto\x12\x14\n\x0clengthMeters\x18\x04 \x01(\x01\">\n\x11PathFeaturesProto\x12)\n\x05paths\x18\x01 \x03(\x0b\x32\x1a.features.PathFeatureProto\"\x89\x01\n\x10\x41reaFeatureProto\x12$\n\x02id\x18\x01 \x01(\x0b\x32\x18.features.FeatureIDProto\x12 \n\x04tags\x18\x02 \x03(\x0b\x32\x12.features.TagProto\x12-\n\x08\x66\x65\x61tures\x18\x03 \x03(\x0b\x32\x1b.features.PathFeaturesProto\"I\n\x13RelationMemberProto\x12$\n\x02id\x18\x01 \x01(\x0b\x32\x18.features.FeatureIDProto\x12\x0c\n\x04role\x18\x03 \x01(\t\"\x8e\x01\n\x14RelationFeatureProto\x12$\n\x02id\x18\x01 \x01(\x0b\x32\x18.features.FeatureIDProto\x12 \n\x04tags\x18\x02 \x03(\x0b\x32\x12.features.TagProto\x12.\n\x07members\x18\x03 \x03(\x0b\x32\x1d.features.RelationMemberProto\"\xd3\x01\n\x0c\x46\x65\x61tureProto\x12,\n\x05point\x18\x01 \x01(\x0b\x32\x1b.features.PointFeatureProtoH\x00\x12*\n\x04path\x18\x02 \x01(\x0b\x32\x1a.features.PathFeatureProtoH\x00\x12*\n\x04\x61rea\x18\x03 \x01(\x0b\x32\x1a.features.AreaFeatureProtoH\x00\x12\x32\n\x08relation\x18\x04 \x01(\x0b\x32\x1e.features.RelationFeatureProtoH\x00\x42\t\n\x07\x66\x65\x61ture*~\n\x0b\x46\x65\x61tureType\x12\x14\n\x10\x46\x65\x61tureTypePoint\x10\x00\x12\x13\n\x0f\x46\x65\x61tureTypePath\x10\x01\x12\x13\n\x0f\x46\x65\x61tureTypeArea\x10\x02\x12\x17\n\x13\x46\x65\x61tureTypeRelation\x10\x03\x12\x16\n\x12\x46\x65\x61tureTypeInvalid\x10\x04\x42\x19Z\x17\x64iagonal.works/b6/protob\x06proto3')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'features_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'Z\027diagonal.works/b6/proto'
  _FEATURETYPE._serialized_start=1104
  _FEATURETYPE._serialized_end=1230
  _TAGPROTO._serialized_start=44
  _TAGPROTO._serialized_end=82
  _FEATUREIDPROTO._serialized_start=84
  _FEATUREIDPROTO._serialized_end=171
  _POINTFEATUREPROTO._serialized_start=174
  _POINTFEATUREPROTO._serialized_end=302
  _PATHFEATUREPROTO._serialized_start=305
  _PATHFEATUREPROTO._serialized_end=464
  _PATHFEATURESPROTO._serialized_start=466
  _PATHFEATURESPROTO._serialized_end=528
  _AREAFEATUREPROTO._serialized_start=531
  _AREAFEATUREPROTO._serialized_end=668
  _RELATIONMEMBERPROTO._serialized_start=670
  _RELATIONMEMBERPROTO._serialized_end=743
  _RELATIONFEATUREPROTO._serialized_start=746
  _RELATIONFEATUREPROTO._serialized_end=888
  _FEATUREPROTO._serialized_start=891
  _FEATUREPROTO._serialized_end=1102
# @@protoc_insertion_point(module_scope)



#!/usr/bin/env python3

import argparse
import json
from operator import ge
import os
import s2sphere
import subprocess
import time
import urllib.request
import unittest

import diagonal_b6 as b6

# TODO: Harmonise these IDs with those in src/diagonal.works/diagonal/test/world.go
COAL_DROPS_YARD_WEST_BUILDING_ID = 222021572
COAL_DROPS_YARD_ENCLOSURE_ID = 500008118
JUBILEE_GREENWAY_ID = 380856
STABLE_STREET_BRIDGE_ID = 140633010
STABLE_STREET_BRIDGE_NORTH_END_ID = 1447052073
STABLE_STREET_BRIDGE_SOUTH_END_ID = 1540349979
VERMUTERIA_NODE_ID = 6082053666
GRANARY_SQUARE_WAY_ID = 222021571
LIGHTERMAN_WAY_ID = 427900370

BUILDINGS_IN_GRANARY_SQUARE = 13
HIGHWAYS_IN_GRANARY_SQUARE = 117
BIKE_PARKING_IN_GRANARY_SQUARE = 11
FOUNTAINS_IN_GRANARY_SQUARE = 4 # Within the pedestrian square itself defined by WKT below not the entire area
STABLE_STREET_BRIDGE_NORTH_END_DEGREE = 7 # After it has been connected

GRANARY_SQUARE_POLYGON_WKT="POLYGON ((-0.1260475 51.5357019,-0.1261001 51.5355674,-0.1261596 51.5354153,-0.1262097 51.535287,-0.1259034 51.5352365,-0.1259462 51.5351347,-0.1255806 51.5350765,-0.1255202 51.5350667,-0.1255004 51.5350372,-0.1254536 51.5349963,-0.1254346 51.5350013,-0.1252611 51.535049,-0.125219 51.5350629,-0.124904 51.5350121,-0.1247915 51.5350326,-0.124709 51.5350541,-0.1247491 51.5351308,-0.1247727 51.5351758,-0.1246766 51.5353808,-0.1246363 51.5354737,-0.125082 51.5355458,-0.1259754 51.5356902,-0.1260475 51.5357019))"
GRANARY_SQUARE_MULTIPOLYGON_WKT="MULTIPOLYGON (((-0.1260475 51.5357019,-0.1261001 51.5355674,-0.1261596 51.5354153,-0.1262097 51.535287,-0.1259034 51.5352365,-0.1259462 51.5351347,-0.1255806 51.5350765,-0.1255202 51.5350667,-0.1255004 51.5350372,-0.1254536 51.5349963,-0.1254346 51.5350013,-0.1252611 51.535049,-0.125219 51.5350629,-0.124904 51.5350121,-0.1247915 51.5350326,-0.124709 51.5350541,-0.1247491 51.5351308,-0.1247727 51.5351758,-0.1246766 51.5353808,-0.1246363 51.5354737,-0.125082 51.5355458,-0.1259754 51.5356902,-0.1260475 51.5357019)))"

EARTH_RADIUS_METERS = 6371.01 * 1000.0

def angle_to_meters(angle):
	return angle.radians * EARTH_RADIUS_METERS

class B6Test(unittest.TestCase):

    def __init__(self, name, connection):
        unittest.TestCase.__init__(self, name)
        self.connection = connection

    def test_find_areas(self):
        names = [building.get_string("name") for (id, building) in self.connection(b6.find_areas(b6.keyed("#building")))]
        self.assertEqual(len(names), BUILDINGS_IN_GRANARY_SQUARE)

    def test_find_point_by_id(self):        
        area = self.connection(b6.find_point(b6.osm_node_id(STABLE_STREET_BRIDGE_SOUTH_END_ID)))
        self.assertEqual(area.id.value, STABLE_STREET_BRIDGE_SOUTH_END_ID)

    def test_find_area_by_id(self):
        area = self.connection(b6.find_area(b6.osm_way_area_id(COAL_DROPS_YARD_WEST_BUILDING_ID)))
        self.assertEqual(area.id.value, COAL_DROPS_YARD_WEST_BUILDING_ID)

    def test_find_non_existant_id(self):
        self.assertEqual(self.connection(b6.find_point(b6.osm_node_id(42))), None)

    def test_find_area_by_wrong_id_type(self):
        with self.assertRaises(Exception): # TODO: Make more specific
            self.connection(b6.find_area(b6.osm_node_id(STABLE_STREET_BRIDGE_SOUTH_END_ID)))

    def test_uk_ons_boundary_id(self):
        self.assertEqual(b6.uk_ons_boundary_id("E01000953").value, 76343044687353)

    def test_relation_members(self):
        greenway = self.connection(b6.find_relation(b6.osm_relation_id(JUBILEE_GREENWAY_ID)))
        paths = ([str(m) for m in greenway.members() if m.is_path()])
        self.assertGreater(len(paths), 10)
        self.assertLess(len(paths), 800)

    def test_or_query(self):
        names = [amenity.get_string("name") for (id, amenity) in self.connection(b6.find(b6.tagged("#amenity", "restaurant").or_(b6.tagged("#amenity", "cafe"))))]
        self.assertIn("Le Cafe Alain Ducasse", names)

    def test_point_degree(self):
        self.assertEqual(self.connection(b6.find_point(b6.osm_node_id(STABLE_STREET_BRIDGE_NORTH_END_ID)).degree()), STABLE_STREET_BRIDGE_NORTH_END_DEGREE)
        degrees = [degree for (id, degree) in self.connection(b6.find_points(b6.within_cap(b6.ll(51.535241, -0.124364), 100)).degree())]
        for d in degrees:
            self.assertGreaterEqual(d, 0)
            self.assertLess(d, 10)

    def test_path_lengths(self):
        lengths = [length for (id, length) in self.connection(b6.find_paths(b6.keyed("#highway")).length())]
        self.assertGreater(len(lengths), 0)
        for l in lengths:
            self.assertGreater(l, 0)
            self.assertLess(l, 1000)

    def test_relation_names(self):
        names = [r.get_string("name") for (id, r) in self.connection(b6.find_relations(b6.keyed("#route")))]
        self.assertIn("Jubilee Greenway", names)

    def test_reachable_areas_from_point(self):
        expression = b6.find_point(b6.osm_node_id(STABLE_STREET_BRIDGE_SOUTH_END_ID)).reachable("walk", 200.0, b6.keyed("#amenity")).get_string("name")
        names = set([name for (_, name) in self.connection(expression)])
        self.assertIn("The Lighterman", names)

    def test_reachable_with_distance(self):
        small = b6.find_point(b6.osm_node_id(STABLE_STREET_BRIDGE_SOUTH_END_ID)).reachable("walk", 100.0, b6.keyed("#amenity")).count()
        large = b6.find_point(b6.osm_node_id(STABLE_STREET_BRIDGE_SOUTH_END_ID)).reachable("walk", 200.0, b6.keyed("#amenity")).count()
        self.assertGreater(self.connection(large), self.connection(small))

    def test_paths_to_reach(self):
        expression = b6.find_point(b6.osm_node_id(STABLE_STREET_BRIDGE_SOUTH_END_ID)).paths_to_reach("walk", 200.0, b6.keyed("#amenity"))
        paths = list(self.connection(expression))
        self.assertGreaterEqual(len(paths), 4)
        for path, count in paths:
            self.assertGreaterEqual(count, 1)
            self.assertLess(count, 100)

    def test_closest_from_point(self):
        expression = b6.find_point(b6.osm_node_id(STABLE_STREET_BRIDGE_SOUTH_END_ID)).closest("walk", 1000.0, b6.tagged("#amenity", "pub"))
        pub = self.connection(expression)
        self.assertEqual("The Lighterman", pub.get_string("name"))

    def test_closest_from_point_distance(self):
        expression = b6.find_point(b6.osm_node_id(STABLE_STREET_BRIDGE_SOUTH_END_ID)).closest_distance("walk", 1000.0, b6.tagged("#amenity", "pub"))
        distance = self.connection(expression)
        self.assertGreater(distance, 100.0)
        self.assertLess(distance, 105.0)

    def test_closest_from_area(self):
        expression = b6.find_area(b6.osm_way_area_id(COAL_DROPS_YARD_WEST_BUILDING_ID)).closest("walk", 1000.0, b6.tagged("#amenity", "pub"))
        pub = self.connection(expression)
        self.assertEqual("The Lighterman", pub.get_string("name"))

    def test_closest_from_point_non_existant(self):
        expression = b6.find_point(b6.osm_node_id(STABLE_STREET_BRIDGE_SOUTH_END_ID)).closest("walk", 1000.0, b6.tagged("#amenity", "nonexistant"))
        self.assertEqual(None, self.connection(expression))

    def test_containing_areas_from_point(self):
        expression = b6.find_point(b6.osm_node_id(STABLE_STREET_BRIDGE_SOUTH_END_ID)).reachable_points("walk", 1000.0, b6.all()).containing_areas(b6.keyed("#shop")).get_string("name")
        names = set([name for (_, name) in self.connection(expression)])
        self.assertIn("Coal Drops Yard", names)

    def test_containing_areas_from_area(self):
        areas = self.connection(b6.find_area(b6.osm_way_area_id(COAL_DROPS_YARD_WEST_BUILDING_ID)).reachable_points("walk", 1000.0, b6.all()).containing_areas(b6.all()))
        self.assertGreater(len(areas), 10)

    def test_count_features(self):
        self.assertEqual(self.connection(b6.find_points(b6.tagged("#amenity", "bicycle_parking")).count()), BIKE_PARKING_IN_GRANARY_SQUARE)
        self.assertEqual(self.connection(b6.find_paths(b6.keyed("#highway")).count()), HIGHWAYS_IN_GRANARY_SQUARE)
        self.assertEqual(self.connection(b6.find_areas(b6.keyed("#building")).count()), BUILDINGS_IN_GRANARY_SQUARE)

    def test_divide_count_features(self):
        self.assertAlmostEqual(self.connection(b6.find_points(b6.tagged("#amenity", "bicycle_parking")).count().divide(10.0)), BIKE_PARKING_IN_GRANARY_SQUARE / 10.0)

    def test_filter(self):
        filtered = self.connection(b6.find_areas(b6.keyed("#amenity")).filter(lambda a: b6.has_key(a, "addr:postcode")))
        self.assertGreater(len(filtered), 0)
        for (_, feature) in filtered:
            self.assertNotEqual(feature.get_string("addr:postcode"), "")

    def test_to_geojson_collection(self):
        geojson = self.connection(b6.to_geojson_collection(b6.find_areas(b6.keyed("#building"))))
        self.assertGreater(len(geojson["features"]), 4)
        for feature in geojson["features"]:
            self.assertIn("#building", feature["properties"])

    def test_to_geojson_with_feature(self):
        closest = b6.find_point(b6.osm_node_id(STABLE_STREET_BRIDGE_SOUTH_END_ID)).closest("walk", 1000.0, b6.tagged("#amenity", "pub"))
        geojson = self.connection(b6.to_geojson(closest))
        self.assertEqual(geojson["type"], "Feature")

    def test_to_geojson_with_missing_feature(self):
        geojson = self.connection(b6.to_geojson(b6.find_point(b6.osm_node_id(1))))
        self.assertEqual(len(geojson["features"]), 0)

    def test_search_within_wkt_polygon(self):
        granarySquare = b6.wkt(GRANARY_SQUARE_POLYGON_WKT)
        self.assertEqual(self.connection(b6.find_areas(b6.tagged("#amenity", "fountain").and_(b6.intersecting(granarySquare))).count()), FOUNTAINS_IN_GRANARY_SQUARE)

    def test_search_within_wkt_multipolygon(self):
        granarySquare = b6.wkt(GRANARY_SQUARE_MULTIPOLYGON_WKT)
        self.assertEqual(self.connection(b6.find_areas(b6.tagged("#amenity", "fountain").and_(b6.intersecting(granarySquare))).count()), FOUNTAINS_IN_GRANARY_SQUARE)

    def test_add_tags(self):
        applied = self.connection(b6.add_tags(b6.find_areas(b6.keyed("#building")).map(lambda building: b6.tag("diagonal:colour", building.get_string("building:levels")))))
        self.assertEqual(len(applied), BUILDINGS_IN_GRANARY_SQUARE)

    def test_add_tags_with_filter(self):
        applied = self.connection(b6.add_tags(b6.find_paths(b6.tagged("#highway", "footway")).filter(lambda h: b6.has_key(h, "bicycle")).map(lambda h: b6.tag("#bicycle", h.get_string("bicycle")))))
        self.assertGreater(len(applied), 0)
        self.assertEqual(self.connection(b6.find_paths(b6.keyed("#bicycle")).count()), len(applied))

    def test_search_for_newly_added_tag(self):
        reachable = b6.find_point(b6.osm_node_id(STABLE_STREET_BRIDGE_SOUTH_END_ID)).reachable("walk", 1000.0, b6.keyed("#amenity"))
        modified = self.connection(b6.add_tags(reachable.map(lambda building: b6.tag("#reachable", "yes"))))
        self.assertGreater(len(modified), 1)
        self.assertEqual(self.connection(b6.find(b6.keyed("#reachable")).count()), len(modified))

    def test_sample_points_along_path(self):
        count = len(self.connection(b6.find_path(b6.osm_way_id(STABLE_STREET_BRIDGE_ID)).sample_points(1.0)))
        self.assertGreater(count, 20)
        self.assertLess(count, 40)

    def test_sample_points_along_paths(self):
        count = 0
        center = s2sphere.LatLng.from_degrees(51.53539, -0.12537)
        for _, ll in self.connection(b6.find_paths(b6.keyed("#highway")).sample_points_along_paths(20.0)):
            d = s2sphere.LatLng.from_degrees(ll[0], ll[1]).get_distance(center)
            self.assertLess(angle_to_meters(d), 500.0)
            count += 1
        self.assertGreater(count, 300)
        self.assertLess(count, 350)

    def test_sightline(self):
        a1 = self.connection(b6.ll(51.53557, -0.12585).sightline(250.0).area())
        a2 = self.connection(b6.ll(51.53557, -0.12585).cap_polygon(250.0).area())
        self.assertGreater(a1/a2, 0.20)
        self.assertLess(a1/a2, 0.30)

    def test_sightline_geojson(self):
        geojson = self.connection(b6.to_geojson(b6.ll(51.53557, -0.12585).sightline(250.0)))
        self.assertEqual(geojson["type"], "Feature")
        self.assertEqual(geojson["geometry"]["type"], "Polygon")
        bridge = b6.osm_way_id(STABLE_STREET_BRIDGE_ID)
        geojson = self.connection(b6.to_geojson_collection(b6.find_path(bridge).sample_points(5.0).map(lambda point: b6.sightline(point, 250.0))))
        self.assertGreater(len(geojson["features"]), 5)
        self.assertLess(len(geojson["features"]), 10)

    def test_s2_points(self):
        count = len(self.connection(b6.find_area(b6.osm_way_area_id(GRANARY_SQUARE_WAY_ID)).s2_points(21, 21)))
        self.assertGreater(count, 400)
        self.assertLess(count, 500)

    def test_rectangle_polygon(self):
        area = self.connection(b6.rectangle_polygon(b6.ll(51.5146, -0.1140), b6.ll(51.5124, -0.0951)).area())
        self.assertGreater(area, 300000)
        self.assertLess(area, 400000)

    def test_s2_grid(self):
        topLeft = (51.5146, -0.1140)
        bottomRight = (51.5124, -0.0951)
        grid = self.connection(b6.rectangle_polygon(b6.ll(*topLeft), b6.ll(*bottomRight)).s2_grid(21))
        rect = s2sphere.LatLngRect.from_point_pair(s2sphere.LatLng.from_degrees(*topLeft), s2sphere.LatLng.from_degrees(*bottomRight))
        for token, _ in grid:
            cell = s2sphere.Cell(s2sphere.CellId.from_token(token))
            self.assertEqual(cell.level(), 21)
            self.assertTrue(rect.intersects(cell.get_rect_bound()))

    def test_s2_covering(self):
        topLeft = (51.5146, -0.1140)
        bottomRight = (51.5124, -0.0951)
        covering = self.connection(b6.rectangle_polygon(b6.ll(*topLeft), b6.ll(*bottomRight)).s2_covering(0, 30))
        rect = s2sphere.LatLngRect.from_point_pair(s2sphere.LatLng.from_degrees(*topLeft), s2sphere.LatLng.from_degrees(*bottomRight))
        count = 0
        for token, _ in covering:
            count += 1
            cell = s2sphere.Cell(s2sphere.CellId.from_token(token))
            self.assertTrue(rect.intersects(cell.get_rect_bound()))
        self.assertLess(count, 10)

    def test_s2_center(self):
        # TODO: Rewrite without using string() once we've sorted out node types
        (lat, lng) = self.connection(b6.s2_center("487604b4fbdc"))
        ll = s2sphere.LatLng.from_degrees(lat, lng)
        expected = s2sphere.LatLng.from_degrees(51.5126733, -0.1140124)
        self.assertLess(ll.get_distance(expected).radians, 0.000001)

    def test_join_paths(self):
        a = self.connection(b6.find_path(b6.osm_way_id(377974549)))
        b = self.connection(b6.find_path(b6.osm_way_id(834245629)))
        joined = self.connection(b6.find_path(b6.osm_way_id(377974549)).join(b6.find_path(b6.osm_way_id(834245629))))
        self.assertLess(abs((joined.length_meters() / (a.length_meters() + b.length_meters())) - 1.0), 0.0001)

    def test_ordered_join_paths(self):
        a = self.connection(b6.find_path(b6.osm_way_id(377974549)))
        b = self.connection(b6.find_path(b6.osm_way_id(834245629)))
        joined = self.connection(b6.find_path(b6.osm_way_id(377974549)).ordered_join(b6.find_path(b6.osm_way_id(834245629))))
        self.assertLess(abs((joined.length_meters() / (a.length_meters() + b.length_meters())) - 1.0), 0.0001)

    def test_points(self):
        points = self.connection(b6.find_path(b6.osm_way_id(STABLE_STREET_BRIDGE_ID)).points())
        expected = s2sphere.LatLng.from_degrees(51.535035, -0.1247934)
        ll = s2sphere.LatLng.from_degrees(*points[0][1])
        self.assertLess(ll.get_distance(expected).radians, 0.000001)

    def test_point_features(self):
        points = self.connection(b6.find_path(b6.osm_way_id(STABLE_STREET_BRIDGE_ID)).point_features())
        self.assertEqual(len(points), 2)
        self.assertEqual(b6.osm_node_id(STABLE_STREET_BRIDGE_NORTH_END_ID), points[0][0])

    def test_paths_by_point(self):
        paths = self.connection(b6.find_point(b6.osm_node_id(STABLE_STREET_BRIDGE_NORTH_END_ID)).point_paths())
        self.assertIn(b6.osm_way_id(STABLE_STREET_BRIDGE_ID), [id for (id, _) in paths])

    def test_interpolate(self):
        (lat, lng) = self.connection(b6.find_path(b6.osm_way_id(377974549)).interpolate(0.5))
        ll = s2sphere.LatLng.from_degrees(lat, lng)
        expected = s2sphere.LatLng.from_degrees(51.5361869, -0.1258445)
        self.assertLess(ll.get_distance(expected).radians, 0.000001)

    def test_snap_area_edges(self):
        building = b6.find_area(b6.osm_way_area_id(COAL_DROPS_YARD_WEST_BUILDING_ID))
        original_area = self.connection(building.area())
        snapped_area = self.connection(b6.snap_area_edges(building, b6.keyed("#highway"), 40.0).area())
        self.assertGreater(snapped_area, original_area)

    def test_geojson_map_areas(self):
        building = b6.find_area(b6.osm_way_area_id(COAL_DROPS_YARD_WEST_BUILDING_ID))
        original_area = self.connection(building.area())
        snapped = b6.to_geojson(building).map_geometries(b6.apply_to_area(lambda a: b6.snap_area_edges(a, b6.keyed("#highway"), 40.0)))
        areas = snapped.geojson_areas().map(lambda a: a.area())
        _, snapped_area = self.connection(areas)[0]
        self.assertGreater(snapped_area, original_area)

    def test_distance_to_point_meters(self):
        distance = self.connection(b6.find_path(b6.osm_way_id(377974549)).distance_to_point_meters(b6.ll(51.53586, -0.12564)))
        self.assertGreater(distance, 24.0)
        self.assertLess(distance, 25.0)

    def test_centroid(self):
        expected = b6.ll(51.5352611, -0.1243803)
        distance = self.connection(b6.distance_meters(b6.find_area(b6.osm_way_area_id(LIGHTERMAN_WAY_ID)).centroid(), expected))
        self.assertLess(distance, 0.1)

    def test_centroids(self):
        origin = b6.ll(51.5352611, -0.1243803)
        distances = self.connection(b6.find_areas(b6.keyed("#building")).centroid().map(lambda p: p.distance_meters(origin)))
        for f, distance in distances:
            self.assertLess(distance, 1000.0)

    def test_find_building_intersecting_point(self):
        point = b6.ll(51.5352611, -0.1243803)
        features = self.connection(b6.find(b6.and_(b6.tagged("#building", "yes"), b6.intersecting(point))))
        self.assertIn("The Lighterman", [f.get_string("name") for (_, f) in features])

    def test_map_area(self):
        areas = list(self.connection(b6.find_areas(b6.keyed("#building")).map(lambda building: building.area())))
        self.assertEqual(len(areas), BUILDINGS_IN_GRANARY_SQUARE)
        for (_, area) in areas:
            self.assertGreater(area, 50)
            self.assertLess(area, 10000)

    def test_path_length(self):
        bridge = self.connection(b6.find_path(b6.osm_way_id(STABLE_STREET_BRIDGE_ID)))
        self.assertGreater(bridge.length_meters(), 20.0)
        self.assertLess(bridge.length_meters(), 30.0)

    def test_connect_points(self):
        modified = self.connection(b6.find_point(b6.osm_node_id(VERMUTERIA_NODE_ID)).connect(b6.find_point(b6.osm_node_id(STABLE_STREET_BRIDGE_NORTH_END_ID))))
        self.assertEqual(len(modified), 1)
        for id in modified.values():
            self.assertTrue(id.is_path())
            self.assertEqual(id.namespace, b6.NAMESPACE_DIAGONAL_ACCESS_POINTS)

    def test_connect_point_to_network(self):
        modified = self.connection(b6.find_point(b6.osm_node_id(VERMUTERIA_NODE_ID)).connect_to_network())
        # The attempt to connect the point to the network fails, as the sample area
        # used by the tests is too small for us to consider any street as being part
        # of the street network, but the test at least verifies the wrapping of
        # network connection code
        self.assertEqual(len(modified), 0)

    def test_connect_area_to_network(self):
        modified = self.connection(b6.find_area(b6.osm_way_area_id(LIGHTERMAN_WAY_ID)).connect_to_network())
        self.assertEqual(len(modified), 0)

    def test_import_geojson_point(self):
        g = {
            "type": "FeatureCollection",
            "features": [
                {
                    "type": "Feature",
                    "geometry": {
                        "type": "Point",
                        "coordinates": [-0.1249292, 51.5352547],
                    },
                    "properties": {
                        "name": "Ruby Violet Truck",
                    }
                }
            ]
        }
        ids = self.connection(b6.import_geojson(b6.parse_geojson(json.dumps(g)), "diagonal.works/test"))
        self.assertEqual(len(ids), 1)
        id = list(ids.values())[0]
        self.assertEqual(self.connection(b6.find_point(id).get_string("name")), "Ruby Violet Truck")

    def test_import_geojson_path(self):
        g = {
            "type": "FeatureCollection",
            "features": [
                {
                    "type": "Feature",
                    "geometry": {
                        "type": "LineString",
                        "coordinates": [[-0.1251651, 51.5349089], [-0.1251580, 51.5347263]],
                    },
                    "properties": {
                        "bridge": "yes",
                    }
                }
            ]
        }
        ids = self.connection(b6.import_geojson(b6.parse_geojson(json.dumps(g)), "diagonal.works/test"))
        self.assertEqual(len(ids), 1)
        id = list(ids.values())[0]
        self.assertEqual(self.connection(b6.find_path(id)).get_string("bridge"), "yes")

    def test_import_geojson_polygon(self):
        g = {
            "type": "FeatureCollection",
            "features": [
                {
                    "type": "Feature",
                    "geometry": {
                        "type": "Polygon",
                        "coordinates": [[[-0.1243817, 51.5354124], [-0.1243411, 51.5351416], [-0.1242415, 51.5353736]]],
                    },
                    "properties": {
                        "building": "yes",
                    }
                }
            ]
        }
        ids = self.connection(b6.import_geojson(b6.parse_geojson(json.dumps(g)), "diagonal.works/test"))
        self.assertEqual(len(ids), 1)
        id = list(ids.values())[0]
        self.assertEqual(self.connection(b6.find_area(id)).get_string("building"), "yes")

    def test_import_geojson_multipolygon(self):
        g = {
            "type": "FeatureCollection",
            "features": [
                {
                    "type": "Feature",
                    "geometry": {
                        "type": "MultiPolygon",
                        "coordinates": [
                            [[[-0.1243817, 51.5354124], [-0.1243411, 51.5351416], [-0.1242415, 51.5353736]]],
                            [[[-0.1239823, 51.5358407], [-0.1240998, 51.5355521], [-0.1238063, 51.5358096]]],
                        ],
                    },
                    "properties": {
                        "building": "yes",
                    }
                }
            ]
        }
        ids = self.connection(b6.import_geojson(b6.parse_geojson(json.dumps(g)), "diagonal.works/test"))
        self.assertEqual(len(ids), 1)
        id = list(ids.values())[0]
        self.assertEqual(self.connection(b6.find_area(id)).get_string("building"), "yes")

    def test_import_geojson_file(self):
        ids = self.connection(b6.import_geojson_file("data/tests/granary-square.geojson", "diagonal.works/test"))
        self.assertGreater(len(ids), 0)
        id = list(ids.values())[0]
        self.assertGreater(self.connection(b6.find_area(id).area()), 100.0)

    def test_evaluate_with_changed_world(self):
        close_road = b6.remove_tag(b6.osm_way_id(STABLE_STREET_BRIDGE_ID), "#highway")
        reachable = b6.find_point(b6.osm_node_id(STABLE_STREET_BRIDGE_SOUTH_END_ID)).reachable("walk", 200.0, b6.keyed("#amenity")).get_string("name")
        before = len(self.connection(reachable))
        after = len(self.connection(b6.with_change(close_road, lambda: reachable)))
        self.assertGreater(before, after)

def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("--http-port", default="10080", help="Host and port on which to serve HTTP")
    parser.add_argument("--grpc-port", default="10081", help="Host and port on which to serve GRPC")
    parser.add_argument("--platform", default=None, help="Current platform, eg darwin/x86_64")
    parser.add_argument("tests", type=str, nargs="*")
    args = parser.parse_args()
    http_address = "localhost:%d" % int(args.http_port)
    grpc_address = "localhost:%d" % int(args.grpc_port)
    platform = args.platform or os.environ.get("TARGETPLATFORM", None) or ("%s/%s" % (os.uname().sysname.lower(), os.uname().machine))
    p = subprocess.Popen([
        "bin/%s/b6" % platform,
        "--http=%s" % http_address,
        "--grpc=%s" % grpc_address,
        "--world=osm:data/tests/granary-square.osm.pbf"
    ])

    ready = False
    for _ in range(20):
        if p.poll() is not None:
            print("Server exited with status %d" % p.poll())
            return
        try:
            r = urllib.request.urlopen("http://%s/healthy" % http_address)
            if r.status == 200:
                if r.read().strip() == b"ok":
                    ready = True
                break
        except urllib.error.URLError:
            pass
        time.sleep(0.5)
    if not ready:
        print("Server was never healthy")
        return
    connection = b6.connect_insecure(grpc_address)
    suite = unittest.TestSuite()
    if len(args.tests) > 0:
        for test in args.tests:
            suite.addTest(B6Test(test, connection))
    else:
        for method in dir(B6Test):
            if method.startswith("test_"):
                suite.addTest(B6Test(method, connection))
    runner = unittest.TextTestRunner()
    runner.run(suite)
    p.terminate()
    p.wait()

if __name__ == "__main__":
    main()


from diagonal_b6 import expression

from diagonal_b6 import api_pb2
from diagonal_b6 import features_pb2

FEATURE_TYPE_POINT = features_pb2.FeatureType.FeatureTypePoint
FEATURE_TYPE_PATH = features_pb2.FeatureType.FeatureTypePath
FEATURE_TYPE_AREA = features_pb2.FeatureType.FeatureTypeArea
FEATURE_TYPE_RELATION = features_pb2.FeatureType.FeatureTypeRelation

NAMESPACE_OSM_NODE = "openstreetmap.org/node"
NAMESPACE_OSM_WAY = "openstreetmap.org/way"
NAMESPACE_OSM_RELATION = "openstreetmap.org/relation"
NAMESPACE_UK_ONS_BOUNDARIES = "statistics.gov.uk/datasets/regions"
NAMESPACE_DIAGONAL_ACCESS_POINTS = "diagonal.works/ns/access-point"

class FeatureID(expression.Literal):

    def __init__(self, type, namespace, value):
        self.type = type
        self.namespace = namespace
        self.value = value

    def is_point(self):
        return self.type == FEATURE_TYPE_POINT

    def is_path(self):
        return self.type == FEATURE_TYPE_PATH
    
    def is_area(self):
        return self.type == FEATURE_TYPE_AREA

    def is_relation(self):
        return self.type == FEATURE_TYPE_RELATION

    def to_literal_proto(self):
        l = api_pb2.LiteralNodeProto()
        l.featureIDValue.type = self.type
        l.featureIDValue.namespace = self.namespace
        l.featureIDValue.value = self.value
        return l

    def __str__(self):
        type = features_pb2.FeatureType.Name(self.type).replace("FeatureType", "").lower()
        return "%s/%s/%d" % (type, self.namespace, self.value)

    def __repr__(self):
        return str(self)

    def __eq__(self, other):
        return self.type == other.type and self.namespace == other.namespace and self.value == other.value

    def __hash__(self):
        return hash(self.type) ^ hash(self.namespace) ^ hash(self.value)

    def _fill_query(self, query):
        query.spatial.area.id.type = self.type
        query.spatial.area.id.namespace = self.namespace
        query.spatial.area.id.value = self.value

class Feature:

    def is_point(self):
        return self.id.is_point()

    def is_path(self):
        return self.id.is_path()
    
    def is_area(self):
        return self.id.is_area()

    def is_relation(self):
        return self.id.is_relation()

    def get(self, key):
        for tag in self._pb.tags:
            if tag.key == key:
                return (key, tag.value)
        return (None, None)

    def get_string(self, key):
        _, value = self.get(key)
        if value is not None:
            return value
        return ""

    def get_int(self, key):
        _, value = self.get(key)
        if value is not None:
            try:
                return int(value)
            except:
                pass
        return 0

    def get_float(self, key):
        _, value = self.get(key)
        if value is not None:
            try:
                return float(value)
            except:
                pass
        return 0

    def all_tags(self):
        return [(tag.key, tag.value) for tag in self._pb.tags]

    def __str__(self):
        type = features_pb2.FeatureType.Name(self.id.type).replace("FeatureType", "").title()
        namespace = features_pb2.FeatureIDProto.Namespace.Name(self.id.namespace).replace("Namespace", "").lower()
        return "<%s %s:%d>" % (type, namespace, self.id.value)

    def _fill_query(self, query):
        return self.id._fill_query(query)

class PointFeature(Feature):

    def __init__(self, p):
        self.id = from_id_proto(p.point.id)
        self._pb = p.point

class PathFeature(Feature):

    def __init__(self, pb):
        self.id = from_id_proto(pb.path.id)
        self._pb = pb.path

    def length_meters(self):
        return self._pb.lengthMeters

class AreaFeature(Feature):

    def __init__(self, pb):
        self.id = from_id_proto(pb.area.id)
        self._pb = pb.area

class RelationFeature(Feature):

    def __init__(self, pb):
        self.id = from_id_proto(pb.relation.id)
        self._pb = pb.relation

    def members(self):
        return [_from_relation_member_proto(m) for m in self._pb.members]

class RelationMember:

    def __init__(self, id, role=None):
        self.id = id
        self.role = role

    def is_point(self):
        return self.id.is_point()

    def is_path(self):
        return self.id.is_path()
    
    def is_area(self):
        return self.id.is_area()

    def is_relation(self):
        return self.id.is_relation()

    def __str__(self):
        return "<RelationMember %s" % (str(self.id),)

def from_id_proto(p):
    return FeatureID(p.type, p.namespace, p.value)

expression.register_literal("featureIDValue", from_id_proto)

def from_applied_change_proto(change):
    applied = {}
    for i in range(0, len(change.original)):
        applied[from_id_proto(change.original[i])] = from_id_proto(change.modified[i])
    return applied

expression.register_literal("appliedChangeValue", from_applied_change_proto)

def id_to_proto(id):
    pb = features_pb2.FeatureIDProto()
    pb.type = id.type
    pb.namespace = id.namespace
    pb.value = id.value
    return pb

def _from_point_proto(p):
    return PointFeature(p)

def _from_path_proto(p):
    return PathFeature(p)

def _from_area_proto(p):
    return AreaFeature(p)

def _from_relation_proto(p):
    return RelationFeature(p)

def _from_relation_member_proto(p):
    return RelationMember(from_id_proto(p.id), p.role)

def from_proto(p):
    oneof = p.WhichOneof("feature")
    if oneof == "point":
        return _from_point_proto(p)
    elif oneof == "path":
        return _from_path_proto(p)
    elif oneof == "area":
        return _from_area_proto(p)
    elif oneof == "relation":
        return _from_relation_proto(p)
    elif oneof == None:
        return None
    raise Exception("Unexpected feature %s" % (p,))

expression.register_literal("featureValue", from_proto)

def osm_node_id(id):
    return FeatureID(FEATURE_TYPE_POINT, NAMESPACE_OSM_NODE, id)

def osm_way_id(id):
    return FeatureID(FEATURE_TYPE_PATH, NAMESPACE_OSM_WAY, id)

def osm_way_area_id(id):
    return FeatureID(FEATURE_TYPE_AREA, NAMESPACE_OSM_WAY, id)

def osm_relation_area_id(id):
    return FeatureID(FEATURE_TYPE_AREA, NAMESPACE_OSM_RELATION, id)

def osm_relation_id(id):
    return FeatureID(FEATURE_TYPE_RELATION, NAMESPACE_OSM_RELATION, id)

def uk_ons_boundary_id(id, year=2011):
    # See GBONS2011IDStrategy in src/diagonal.works/b6/ingest/gdal/source.go
    if len(id) != 9:
        raise "Expected a string of 9 characters"
    codeBits = ord(id[0]) << 40
    yearBits = (year-1900) << 32
    return FeatureID(FEATURE_TYPE_AREA, NAMESPACE_UK_ONS_BOUNDARIES, codeBits|yearBits|int(id[1:]))


from diagonal_b6 import expression

from diagonal_b6 import api_pb2
from diagonal_b6 import geometry_pb2

def from_point_proto(p):
    return (float(p.lat_e7) / 1e7, float(p.lng_e7) / 1e7)

expression.register_literal("pointValue", from_point_proto)

def from_polyline_proto(p):
    return Polyline(p)

expression.register_literal("pathValue", from_polyline_proto)

def from_multipolygon_proto(p):
    return MultiPolygon(p)

expression.register_literal("areaValue", from_multipolygon_proto)

class Polyline(expression.Node):

    def __init__(self, pb):
        self._pb = pb

    def length_meters(self):
        return self._pb.length_meters

    def to_node_proto(self):
        n = api_pb2.NodeProto()
        n.literal.pathValue.CopyFrom(self._pb)
        return n

    def __str__(self):
        return str(self._pb)

class Polygon(expression.Node):

    def __init__(self, pb):
        self._pb = pb

    def to_node_proto(self):
        n = api_pb2.NodeProto()
        n.literal.areaValue.polygons.add().CopyFrom(self._pb)
        return n

    def __str__(self):
        return str(self._pb)

class MultiPolygon(expression.Node):

    def __init__(self, pb):
        self._pb = pb

    def to_node_proto(self):
        n = api_pb2.NodeProto()
        n.literal.areaValue.CopyFrom(self._pb)
        return n

    def __str__(self):
        return str(self._pb)

def wkt(s):
    if s.startswith("POLYGON "):
        return _wkt_polygon(s)
    elif s.startswith("MULTIPOLYGON "):
        return _wkt_multipolygon(s)
    raise ValueError("Can't parse WKT %s" % (repr(s),))

def _wkt_polygon(s):
    header = "POLYGON "
    polygon_wkt = s[len(header):]
    polygon = geometry_pb2.PolygonProto()
    _fill_polygon_from_wkt(polygon, polygon_wkt)
    return Polygon(polygon)

def _wkt_multipolygon(s):
    header = "MULTIPOLYGON "
    multipolygon_wkt = s[len(header):]
    multipolygon = geometry_pb2.MultiPolygonProto()
    for group in _parse_wkt_groups(multipolygon_wkt):
        _fill_polygon_from_wkt(multipolygon.polygons.add(), group)
    return MultiPolygon(multipolygon)

def _fill_polygon_from_wkt(polygon, wkt):
    for group in _parse_wkt_groups(wkt):
        _fill_loop_from_wkt(polygon.loops.add(), group)

def _fill_loop_from_wkt(loop, wkt):
    for (lat, lng) in _parse_wkt_points(wkt):
        point = loop.points.add()
        point.lat_e7 = int(lat * 1e7)
        point.lng_e7 = int(lng * 1e7)

def _parse_wkt_groups(s):
    if s[0] != "(" and s[len(s)-1] != ")":
        raise ValueError("Expected a bracketed WKT group: %s" % (repr(s),))
    groups = []
    i = 1
    while i < len(s) - 1:        
        if s[i] == "(":
            depth = 1
            start = i
            i += 1
            while True:
                if i >= len(s) - 1:
                    raise ValueError("Expected closing bracket in group %s" % (repr(s),))
                if s[i] == ")":
                    depth -= 1
                    if depth == 0:
                        i += 1
                        break
                if s[i] == "(":
                    depth += 1
                i += 1
            groups.append(s[start:i])
        elif s[i] == " " or s[i] == ",":
            i += 1
        else:
            raise ValueError("Invalid character %s in group %s" % (repr(s[i]), repr(s)))
    return groups

def _parse_wkt_points(s):
    if s[0] != "(" and s[len(s)-1] != ")":
        raise ValueError("Expected a bracketed WKT group of points %s" % (repr(s),))
    parsed = []
    for coordinates in s[1:-1].split(","):
        cs = [float(c) for c in coordinates.split(" ") if len(c) > 0]
        if len(cs) != 2:
            raise ValueError("Expected a coordinate with 2 values, found %s" % (repr(coordinates),))
        parsed.append((cs[1], cs[0]))
    return parsed

